import os
import copy
import json
import importlib
from datetime import datetime
import numpy as np
import emcee
import h5py

import appletree as apt
from appletree import randgen
from appletree import Parameter
from appletree import Likelihood
from appletree.utils import load_json
from appletree.share import _cached_configs, set_global_config

os.environ['OMP_NUM_THREADS'] = '1'


class Context():
    """Combine all likelihood (e.g. Rn220, Ar37),
    handle MCMC and post-fitting analysis
    """

    def __init__(self, config):
        """Create an appletree context

        :param config: dict or str, configuration file name or dictionary
        """
        if isinstance(config, str):
            config = load_json(config)

        # url_base and configs are not mandatory
        if 'url_base' in config.keys():
            self.update_url_base(config['url_base'])

        if 'configs' in config.keys():
            self.set_config(config['configs'])

        self.backend_h5 = config.get('backend_h5', None)

        self.likelihoods = {}

        self.par_config = self.get_parameter_config(config['par_config'])
        self.needed_parameters = self.update_parameter_config(config['likelihoods'])

        self.par_manager = Parameter(self.par_config)

        self.register_all_likelihood(config)

    def __getitem__(self, keys):
        """Get likelihood in context"""
        return self.likelihoods[keys]

    def register_all_likelihood(self, config):
        """Create all appletree likelihoods

        :param config: dict, configuration file name or dictionary
        """
        components = importlib.import_module('appletree.components')

        for key, value in config['likelihoods'].items():
            likelihood = copy.deepcopy(value)

            # update data file path
            data_file_name = likelihood["data_file_name"]
            if not os.path.exists(data_file_name):
                likelihood["data_file_name"] = data_file_name

            self.register_likelihood(key, likelihood)

            for k, v in likelihood['components'].items():
                # dynamically import components
                if isinstance(v, str):
                    self.register_component(key, getattr(components, v), k)
                else:
                    self.register_component(
                        key,
                        getattr(components, v['component_cls']),
                        k,
                        v.get('file_name', None),
                    )

    def register_likelihood(self,
                            likelihood_name,
                            likelihood_config):
        """Create an appletree likelihood

        :param likelihood_name: name of Likelihood
        :param likelihood_config: dict of likelihood configuration
        """
        if likelihood_name in self.likelihoods:
            raise ValueError(f'Likelihood named {likelihood_name} already existed!')
        self.likelihoods[likelihood_name] = Likelihood(
            name=likelihood_name,
            **likelihood_config,
        )

    def register_component(self,
                           likelihood_name,
                           component_cls,
                           component_name,
                           file_name=None):
        """Register component to likelihood

        :param likelihood_name: name of Likelihood
        :param component_cls: class of Component
        :param component_name: name of Component
        """
        self[likelihood_name].register_component(
            component_cls,
            component_name,
            file_name,
        )
        # Update needed parameters
        self.needed_parameters |= self.likelihoods[likelihood_name].needed_parameters

    def print_context_summary(self, short=True):
        """Print summary of the context."""
        self._sanity_check()

        print('\n'+'='*40)
        for key, likelihood in self.likelihoods.items():
            print(f'LIKELIHOOD {key}')
            likelihood.print_likelihood_summary(short=short)
            print('\n'+'='*40)

    def log_posterior(self, parameters, batch_size=1_000_000):
        """Get log likelihood of given parameters

        :param batch_size: int of number of simulated events
        :param parameters: dict of parameters used in simulation
        """
        self.par_manager.set_parameter(parameters)

        key = randgen.get_key()
        log_posterior = 0
        for likelihood in self.likelihoods.values():
            key, log_likelihood_i = likelihood.get_log_likelihood(
                key,
                batch_size,
                self.par_manager.get_all_parameter(),
            )
            log_posterior += log_likelihood_i

        log_prior = self.par_manager.log_prior
        log_posterior += log_prior

        return log_posterior, log_prior

    def _get_backend(self, nwalkers, ndim):
        if self.backend_h5 is None:
            backend = None
            print('With no backend')
        else:
            backend = emcee.backends.HDFBackend(self.backend_h5)
            backend.reset(nwalkers, ndim)
            print(f'With h5 backend {self.backend_h5}')
        return backend

    def fitting(self, nwalkers=200, iteration=500, batch_size=1_000_000):
        """Fitting posterior distribution of needed parameters

        :param nwalkers: int, number of walkers in the ensemble
        :param iteration: int, number of steps to generate
        """
        self._sanity_check()

        p0 = []
        for _ in range(nwalkers):
            self.par_manager.sample_init()
            p0.append(self.par_manager.parameter_fit_array)

        ndim = len(self.par_manager.parameter_fit_array)

        backend = self._get_backend(nwalkers, ndim)
        self.sampler = emcee.EnsembleSampler(
            nwalkers,
            ndim,
            self.log_posterior,
            backend=backend,
            blobs_dtype=np.float32,
            parameter_names=self.par_manager.parameter_fit,
            kwargs = {'batch_size': batch_size},
        )

        result = self.sampler.run_mcmc(
            p0,
            iteration,
            store=True,
            progress=True,
        )

        self._dump_meta()
        return result

    def continue_fitting(self, context, iteration=500, batch_size=1_000_000):
        """Continue a fitting of another context

        :param context: appletree context
        :param iteration: int, number of steps to generate
        """
        # Final iteration
        final_iteration = context.sampler.get_chain()[-1, :, :]
        p0 = final_iteration.tolist()

        ndim = len(self.par_manager.parameter_fit)
        # Init sampler for current context
        backend = self._get_backend(len(p0), ndim)
        self.sampler = emcee.EnsembleSampler(
            len(p0),
            ndim,
            self.log_posterior,
            backend=backend,
            blobs_dtype=np.float32,
            parameter_names=self.par_manager.parameter_fit,
            kwargs = {'batch_size': batch_size},
        )

        result = self.sampler.run_mcmc(
            p0,
            iteration,
            store=True,
            progress=True,
            skip_initial_state_check=True,
        )

        self._dump_meta()
        return result

    def get_post_parameters(self):
        """Get parameters correspondes to max posterior"""
        logp = self.sampler.get_log_prob(flat=True)
        chain = self.sampler.get_chain(flat=True)
        mpe_parameters = chain[np.argmax(logp)]
        mpe_parameters = emcee.ensemble.ndarray_to_list_of_dicts(
            [mpe_parameters],
            self.sampler.parameter_names,
        )[0]
        parameters = copy.deepcopy(self.par_manager.get_all_parameter())
        parameters.update(mpe_parameters)
        return parameters

    def get_all_post_parameters(self, **kwargs):
        """Return all posterior parameters"""
        chain = self.sampler.get_chain(**kwargs)
        return chain

    def dump_post_parameters(self, file_name):
        """Dump max posterior parameter in .json file"""
        parameters = self.get_post_parameters()
        with open(file_name, 'w') as fp:
            json.dump(parameters, fp)

    def _dump_meta(self, metadata=None):
        """Save parameters name as attributes"""
        if metadata is None:
            metadata = {
                'version': apt.__version__,
                'date': datetime.now().strftime('%Y%m%d_%H:%M:%S'),
            }
        if self.backend_h5 is not None:
            name = self.sampler.backend.name
            with h5py.File(self.backend_h5, 'r+') as opt:
                opt[name].attrs['metadata'] = json.dumps(metadata)
                # parameters prior configuration
                opt[name].attrs['par_config'] = json.dumps(self.par_manager.par_config)
                # max posterior parameters
                opt[name].attrs['post_parameters'] = json.dumps(self.get_post_parameters())
                # the order of parameters saved in backend
                opt[name].attrs['parameter_fit'] = self.par_manager.parameter_fit
                # instructions
                opt[name].attrs['config'] = json.dumps(self.config)
                # configurations, maybe users will manually add some maps
                opt[name].attrs['_cached_configs'] = json.dumps(_cached_configs)

    def get_template(self,
                     likelihood_name: str,
                     component_name: str,
                     batch_size: int = 1_000_000,
                     seed: int = None):
        """Get parameters correspondes to max posterior

        :param likelihood_name: name of Likelihood
        :param component_name: name of Component
        :param batch_size: int of number of simulated events
        :param seed: random seed
        """
        parameters = self.get_post_parameters()
        key = randgen.get_key(seed=seed)

        key, result = self[likelihood_name][component_name].simulate(
            key,
            batch_size, parameters,
        )
        return result

    def _sanity_check(self):
        """Check if needed parameters are provided."""
        needed = set(self.needed_parameters)
        provided = set(self.par_manager.get_all_parameter().keys())
        # We will not update unneeded parameters!
        if not provided.issubset(needed):
            mes = f'Parameter manager should provide needed parameters only, '
            mes += f'{provided - needed} not needed'
            raise RuntimeError(mes)

    def update_url_base(self, url_base):
        """Update url_base in appletree.share"""
        print(f'Updated url_base to {url_base}')
        set_global_config({'url_base': url_base})

    def get_parameter_config(self, par_config):
        """Get configuration for parameter manager

        :param par_config: str, parameters configuration file
        """
        par_config = load_json(par_config)
        return par_config

    def update_parameter_config(self, likelihoods):
        needed_parameters = set()
        needed_rate_parameters = []
        from_parameters = []
        for likelihood in likelihoods.values():
            for k, v in likelihood['copy_parameters'].items():
                # specify rate scale
                # normalization factor, for AC & ER, etc.
                self.par_config.update({k: self.par_config[v]})
                from_parameters.append(v)
                needed_parameters.add(k)
            for k in likelihood['components'].keys():
                needed_rate_parameters.append(k + '_rate')
        for p in from_parameters:
            if p not in needed_rate_parameters and p in self.par_config:
                # Drop unused parameters
                self.par_config.pop(p)
        return needed_parameters

    def set_config(self, configs):
        """Set new configuration options

        :param configs: dict, configuration file name or dictionary
        """
        if not hasattr(self, 'config'):
            self.config = dict()

        # update configuration only in this Context
        self.config.update(configs)

        # also store required configurations to appletree.share
        set_global_config(configs)

    def lineage(self, data_name: str = 'cs2'):
        """Return lineage of plugins."""
        assert isinstance(data_name, str)
        pass



from warnings import warn
from functools import partial

import numpy as np
import pandas as pd
from jax import numpy as jnp
import pandas as pd

import appletree
from appletree import utils
from appletree.plugin import Plugin
from appletree.share import _cached_configs, _cached_functions, set_global_config
from appletree.utils import exporter, load_data
from appletree.hist import make_hist_mesh_grid, make_hist_irreg_bin_2d

export, __all__ = exporter()


@export
class Component:
    """Base class of component"""

    # Do not initialize this class because it is base
    __is_base = True

    rate_name: str = ''
    norm_type: str = ''
    add_eps_to_hist: bool = True

    def __init__(self,
                 name: str = None,
                 llh_name: str = None,
                 bins: list = [],
                 bins_type: str = '',
                 **kwargs):
        """Initialization.

        :param bins: bins to generate the histogram.

          * For irreg bins_type, bins must be bin edges of the two dimensions.
          * For meshgrid bins_type, bins are sent to jnp.histogramdd.
        :param bins_type: binning scheme, can be either irreg or meshgrid.
        """
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name
        if llh_name is None:
            self.llh_name = self.__class__.__name__ + '_llh'
        else:
            self.llh_name = llh_name
        self.bins = bins
        self.bins_type = bins_type
        self.needed_parameters = set()

        if self.bins_type == 'meshgrid':
            warning = 'The usage of meshgrid binning is highly discouraged.'
            warn(warning)

    def _clip(self, result: list):
        """Clip simulated result"""
        mask = np.ones(len(result[-1]), dtype=bool)
        for i in range(len(result) - 1):
            mask &= result[i] > np.array(self.bins[i]).min()
            mask &= result[i] < np.array(self.bins[i]).max()
        for i in range(len(result)):
            result[i] = result[i][mask]
        return result

    @property
    def _use_mcinput(self):
        return 'Bootstrap' in self._plugin_class_registry['energy'].__name__

    def simulate_hist(self, *args, **kwargs):
        """Hook for simulation with histogram output."""
        raise NotImplementedError

    def multiple_simulations(self, key, batch_size, parameters, times):
        """Simulate many times and
        move results to CPU because the memory limit of GPU
        """
        results_pile = []
        for _ in range(times):
            key, results = self.simulate(key, batch_size, parameters)
            results_pile.append(np.array(results))
        return key, np.hstack(results_pile)

    def multiple_simulations_compile(self, key, batch_size, parameters, times):
        """Simulate many times after new compilation and
        move results to CPU because the memory limit of GPU
        """
        results_pile = []
        for _ in range(times):
            if _cached_configs['g4'] and self._use_mcinput:
                if isinstance(_cached_configs['g4'], dict):
                    g4_file_name = _cached_configs['g4'][self.llh_name][0]
                    _cached_configs['g4'][self.llh_name] = [
                        g4_file_name, batch_size, key.sum().item()]
                else:
                    g4_file_name = _cached_configs['g4'][0]
                    _cached_configs['g4'] = [g4_file_name, batch_size, key.sum().item()]
            self.compile()
            key, results = self.multiple_simulations(key, batch_size, parameters, 1)
            results_pile.append(results)
        return key, np.hstack(results_pile)

    def implement_binning(self, mc, eff):
        """Apply binning to MC data.

        :param mc: data from simulation.
        :param eff: efficiency of each event, as the weight when making a histogram.
        """
        if self.bins_type == 'irreg':
            hist = make_hist_irreg_bin_2d(mc, *self.bins, weights=eff)
        elif self.bins_type == 'meshgrid':
            hist = make_hist_mesh_grid(mc, bins=self.bins, weights=eff)
        else:
            raise ValueError(f'Unsupported bins_type {self.bins_type}!')
        if self.add_eps_to_hist:
            # as an uncertainty to prevent blowing up
            hist = jnp.clip(hist, 1., jnp.inf)
        return hist

    def get_normalization(self, hist, parameters, batch_size=None):
        """Return the normalization factor of the histogram."""
        if self.norm_type == 'on_pdf':
            normalization_factor = 1 / jnp.sum(hist) * parameters[self.rate_name]
        elif self.norm_type == 'on_sim':
            if self._use_mcinput:
                bootstrap_name = self._plugin_class_registry['energy'].__name__
                bootstrap_name = bootstrap_name + '_' + self.name
                n_events_selected = _cached_functions[self.llh_name][bootstrap_name].g4.n_events_selected
                normalization_factor = 1 / n_events_selected * parameters[self.rate_name]
            else:
                normalization_factor = 1 / batch_size * parameters[self.rate_name]
        else:
            raise ValueError(f'Unsupported norm_type {self.norm_type}!')
        return normalization_factor

    def deduce(self, *args, **kwargs):
        """Hook for workflow deduction."""
        raise NotImplementedError

    def compile(self):
        """Hook for compiling simulation code."""
        pass


@export
class ComponentSim(Component):
    """Component that needs MC simulations."""

    # Do not initialize this class because it is base
    __is_base = True

    code: str = None
    old_code: str = None

    def __init__(self,
                 *args, **kwargs):
        """Initialization"""
        super().__init__(*args, **kwargs)
        self._plugin_class_registry = {}

    def register(self, plugin_class):
        """Register a plugin to the component."""
        if isinstance(plugin_class, (tuple, list)):
            # Shortcut for multiple registration
            for x in plugin_class:
                self.register(x)
            return

        if not hasattr(plugin_class, 'provides'):
            # No output name specified: construct one from the class name
            snake_name = appletree.camel_to_snake(plugin_class.__name__)
            plugin_class.provides = (snake_name,)

        # Ensure plugin_class.provides is a tuple
        if isinstance(plugin_class.provides, str):
            plugin_class.provides = tuple([plugin_class.provides])

        for p in plugin_class.provides:
            self._plugin_class_registry[p] = plugin_class

        already_seen = []
        for plugin in self._plugin_class_registry.values():

            if plugin in already_seen:
                continue
            already_seen.append(plugin)

            for config_name, items in plugin.takes_config.items():
                # Looping over the configs of the new plugin and check if
                # they can be found in the already registered plugins:
                for new_config, new_items in plugin_class.takes_config.items():
                    if new_config != config_name:
                        continue
                    if items.default == new_items.default:
                        continue
                    else:
                        mes = f'Two plugins have a different file name'
                        mes += f' for the same config. The config'
                        mes += f' "{new_config}" in "{plugin.__name__}" takes'
                        mes += f' the file name as "{new_items.default}"  while in'
                        mes += f' "{plugin_class.__name__}" the file name'
                        mes += f' is set to "{items.default}". Please change'
                        mes += f' one of the file names.'
                        raise ValueError(mes)

    def register_all(self, module):
        """Register all plugins defined in module.
        Can pass a list/tuple of modules to register all in each.
        """
        if isinstance(module, (tuple, list)):
            # Shortcut for multiple registration
            for x in module:
                self.register_all(x)
            return

        for x in dir(module):
            x = getattr(module, x)
            if not isinstance(x, type(type)):
                continue
            if issubclass(x, Plugin):
                self.register(x)

    def dependencies_deduce(self,
                            data_names: list = ('cs1', 'cs2', 'eff'),
                            dependencies: list = None,
                            nodep_data_name: str = 'batch_size') -> list:
        """Deduce dependencies.

        :param data_names: data names that simulation will output.
        :param dependencies: dependency tree.
        :param nodep_data_name: data_name without dependency will not be deduced
        """
        if dependencies is None:
            dependencies = []

        for data_name in data_names:
            # usually `batch_size` have no dependency
            if data_name == nodep_data_name:
                continue
            try:
                dependencies.append({
                    'plugin': self._plugin_class_registry[data_name],
                    'provides': data_name,
                    'depends_on': self._plugin_class_registry[data_name].depends_on,
                })
            except KeyError:
                raise ValueError(f'Can not find dependency for {data_name}')

        for data_name in data_names:
            # `batch_size` has no dependency
            if data_name == nodep_data_name:
                continue
            dependencies = self.dependencies_deduce(
                data_names=self._plugin_class_registry[data_name].depends_on,
                dependencies=dependencies,
                nodep_data_name=nodep_data_name,
            )

        return dependencies

    def dependencies_simplify(self, dependencies):
        """Simplify the dependencies."""
        already_seen = []
        self.worksheet = []
        self.needed_parameters.add(self.rate_name)
        for plugin in dependencies[::-1]:
            plugin = plugin['plugin']
            if plugin.__name__ in already_seen:
                continue
            self.worksheet.append([plugin.__name__, plugin.provides, plugin.depends_on])
            already_seen.append(plugin.__name__)
            self.needed_parameters |= set(plugin.parameters)

    def flush_source_code(self,
                          data_names: list = ['cs1', 'cs2', 'eff'],
                          func_name: str = 'simulate',
                          nodep_data_name: str = 'batch_size'):
        """Infer the simulation code from the dependency tree."""
        self.func_name = func_name

        if not isinstance(data_names, (list, str)):
            raise RuntimeError(f'data_names must be list or str, but given {type(data_names)}')
        if isinstance(data_names, str):
            data_names = [data_names]

        code = ''
        indent = ' ' * 4

        code += 'from functools import partial\n'
        code += 'from jax import jit\n'

        # import needed plugins
        for work in self.worksheet:
            plugin = work[0]
            code += f'from appletree.plugins import {plugin}\n'

        # initialize new instances
        for work in self.worksheet:
            plugin = work[0]
            instance = plugin + '_' + self.name
            code += f"{instance} = {plugin}('{self.llh_name}')\n"

        # define functions
        code += '\n'
        if nodep_data_name == 'batch_size':
            code += '@partial(jit, static_argnums=(1, ))\n'
        else:
            code += '@jit\n'
        code += f'def {func_name}(key, {nodep_data_name}, parameters):\n'

        for work in self.worksheet:
            provides = 'key, ' + ', '.join(work[1])
            depends_on = ', '.join(work[2])
            instance = work[0] + '_' + self.name
            code += f'{indent}{provides} = {instance}(key, parameters, {depends_on})\n'
        output = 'key, ' + '[' + ', '.join(data_names) + ']'
        code += f'{indent}return {output}\n'

        self.code = code

        if func_name in _cached_functions[self.llh_name].keys():
            warning = f'Function name {func_name} is already cached. '
            warning += 'Running compile() will overwrite it.'
            warn(warning)

    @property
    def code(self):
        """Code of simulation function."""
        return self._code

    @code.setter
    def code(self, code):
        self._code = code
        if self.llh_name not in _cached_functions.keys():
            _cached_functions[self.llh_name] = {}
        self._compile = partial(exec, self.code, _cached_functions[self.llh_name])

    def deduce(self,
               data_names: list = ('cs1', 'cs2'),
               func_name: str = 'simulate',
               nodep_data_name: str = 'batch_size',
               force_no_eff: bool = False):
        """Deduce workflow and code.

        :param data_names: data names that simulation will output.
        :param func_name: name of the simulation function, used to cache it.
        :param nodep_data_name: data_name without dependency will not be deduced
        :param force_no_eff: force to ignore the efficiency, used in yield prediction
        """
        if not isinstance(data_names, (list, tuple)):
            raise ValueError(f'Unsupported data_names type {type(data_names)}!')
        # make sure that 'eff' is the last data_name
        if 'eff' in data_names:
            data_names = list(data_names)
            data_names.remove('eff')
        if not force_no_eff:
            data_names = list(data_names) + ['eff']

        dependencies = self.dependencies_deduce(data_names, nodep_data_name=nodep_data_name)
        self.dependencies_simplify(dependencies)
        self.flush_source_code(data_names, func_name, nodep_data_name)

    def compile(self):
        """Build simulation function and cache it to share._cached_functions."""
        self._compile()
        self.simulate = _cached_functions[self.llh_name][self.func_name]

    def simulate_hist(self,
                      key,
                      batch_size,
                      parameters):
        """Simulate and return histogram.

        :param key: key used for pseudorandom generator.
        :param batch_size: number of events to be simulated.
        :param parameters: a dictionary that contains all parameters needed in simulation.
        """
        key, result = self.simulate(key, batch_size, parameters)
        mc = result[:-1]
        assert len(mc) == len(self.bins), "Length of bins must be the same as length of bins_on!"
        mc = jnp.asarray(mc).T
        eff = result[-1]  # we guarantee that the last output is efficiency in self.deduce

        hist = self.implement_binning(mc, eff)
        normalization_factor = self.get_normalization(hist, parameters, batch_size)
        hist *= normalization_factor

        return key, hist

    def simulate_weighed_data(self,
                              key,
                              batch_size,
                              parameters):
        """Simulate and return histogram."""
        key, result = self.simulate(key, batch_size, parameters)
        # Move data to CPU
        result = [np.array(r) for r in result]
        # Clip data points out of ROI
        result = self._clip(result)
        mc = result[:-1]
        assert len(mc) == len(self.bins), "Length of bins must be the same as length of bins_on!"
        mc = jnp.asarray(mc).T
        eff = jnp.asarray(result[-1])  # we guarantee that the last output is efficiency in self.deduce

        hist = self.implement_binning(mc, eff)
        normalization_factor = self.get_normalization(hist, parameters, batch_size)
        result[-1] *= normalization_factor

        return key, result

    def save_code(self, file_path):
        """Save the code to file."""
        with open(file_path, 'w') as f:
            f.write(self.code)

    def lineage(self, data_name: str = 'cs2'):
        """Return lineage of plugins."""
        assert isinstance(data_name, str)
        pass

    def set_config(self, configs):
        """Set new global configuration options

        :param configs: dict, configuration file name or dictionary
        """
        set_global_config(configs)

    def show_config(self, data_names: list = ('cs1', 'cs2', 'eff')):
        """
        Return configuration options that affect data_names.

        :param data_names: Data type name
        """
        dependencies = self.dependencies_deduce(
            data_names,
            nodep_data_name='batch_size',
        )
        r = []
        seen = []

        for dep in dependencies:
            p = dep['plugin']
            # Track plugins we already saw, so options from
            # multi-output plugins don't come up several times
            if p in seen:
                continue
            seen.append(p)

            for config in p.takes_config.values():
                try:
                    default = config.get_default()
                except:
                    default = appletree.OMITTED
                current = _cached_configs.get(config.name, None)
                if isinstance(current, dict):
                    current = current[self.llh_name]
                r.append(dict(
                    option=config.name,
                    default=default,
                    current=current,
                    applies_to=p.provides,
                    help=config.help))
        if len(r):
            df = pd.DataFrame(r, columns=r[0].keys())
        else:
            df = pd.DataFrame([])

        # Then you can print the dataframe like:
        # straxen.dataframe_to_wiki(df, title=f'{data_names}', float_digits=1)
        return df

    def new_component(self):
        """
        Generate new component with same binning,
        usually used on predicting yields
        """
        component = self.__class__(
            name=self.name + '_copy',
            bins=self.bins,
            bins_type=self.bins_type,
        )
        return component


@export
class ComponentFixed(Component):
    """Component whose shape is fixed."""

    # Do not initialize this class because it is base
    __is_base = True

    def __init__(self,
                 *args, **kwargs):
        """Initialization"""
        if not kwargs.get('file_name', None):
            raise ValueError('Should provide file_name for ComponentFixed!')
        else:
            self._file_name = kwargs.get('file_name', None)
        super().__init__(*args, **kwargs)

    def deduce(self,
               data_names: list = ('cs1', 'cs2')):
        """Deduce the needed parameters and make the fixed histogram."""
        self.data = load_data(self._file_name)[list(data_names)].to_numpy()
        self.eff = jnp.ones(len(self.data))
        self.hist = self.implement_binning(self.data, self.eff)
        self.needed_parameters.add(self.rate_name)

    def simulate(self):
        """Fixed component does not need to simulate."""
        raise NotImplementedError

    def simulate_hist(self,
                      parameters,
                      *args, **kwargs):
        """Return the fixed histogram."""
        normalization_factor = self.get_normalization(self.hist, parameters, len(self.data))
        return self.hist * normalization_factor

    def simulate_weighed_data(self,
                              parameters,
                              *args, **kwargs):
        """Simulate and return histogram."""
        result = [r for r in self.data.T]
        result.append(np.array(self.eff))
        # Clip all simulated data points
        result = self._clip(result)
        normalization_factor = self.get_normalization(self.hist, parameters, len(self.data))
        result[-1] *= normalization_factor

        return result


@export
def add_component_extensions(module1, module2, force=False):
    """Add components of module2 to module1"""
    utils.add_extensions(module1, module2, Component, force=force)


@export
def _add_component_extension(module, component, force=False):
    """Add component to module"""
    utils._add_extension(module, component, Component, force=force)



from warnings import warn

import numpy as np
from jax import numpy as jnp

from appletree.hist import make_hist_mesh_grid, make_hist_irreg_bin_2d
from appletree.utils import load_data, get_equiprob_bins_2d
from appletree.component import Component, ComponentSim, ComponentFixed


class Likelihood:
    """Combine all components (e.g. ER, AC, Wall), and calculate log posterior likelihood"""

    def __init__(self, name: str = None, **config):
        """Create an appletree likelihood

        :param config: Dictionary with configuration options that will be applied, should include:

          * data_file_name: the data used in fitting, usually calibration data
          * bins_type: either meshgrid or equiprob
          * bins_on: observables where we will perform inference on, usually [cs1, cs2]
          * x_clip, y_clip: ROI of the fitting, should be list of upper and lower boundary
        """
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name
        self.components = {}
        self._config = config
        self._data_file_name = config['data_file_name']
        self._bins_type = config['bins_type']
        self._bins_on = config['bins_on']
        self._bins = config['bins']
        self._dim = len(self._bins_on)
        if self._dim != 2:
            raise ValueError('Currently only support 2D fitting')
        self.needed_parameters = set()
        self._sanity_check()

        self.data = load_data(self._data_file_name)[self._bins_on].to_numpy()
        mask = (self.data[:, 0] > config['x_clip'][0])
        mask &= (self.data[:, 0] < config['x_clip'][1])
        mask &= (self.data[:, 1] > config['y_clip'][0])
        mask &= (self.data[:, 1] < config['y_clip'][1])
        self.data = self.data[mask]

        if self._bins_type == 'meshgrid':
            warning = f'The usage of meshgrid binning is highly discouraged.'
            warn(warning)
            self.component_bins_type = 'meshgrid'
            x_bins = jnp.linspace(*config['x_clip'], self._bins[0] + 1)
            y_bins = jnp.linspace(*config['y_clip'], self._bins[1] + 1)
            self._bins = (x_bins, y_bins)
            self.data_hist = make_hist_mesh_grid(
                self.data,
                bins=self._bins,
                weights=jnp.ones(len(self.data)),
            )
        elif self._bins_type == 'equiprob':
            if self._dim != 2:
                raise RuntimeError('only 2D equiprob binned likelihood is supported!')
            self._bins = get_equiprob_bins_2d(
                self.data,
                self._bins,
                x_clip=config['x_clip'],
                y_clip=config['y_clip'],
                which_np=jnp)
            self.component_bins_type = 'irreg'
            self.data_hist = make_hist_irreg_bin_2d(
                self.data,
                bins_x=self._bins[0],
                bins_y=self._bins[1],
                weights=jnp.ones(len(self.data)),
            )
        elif self._bins_type == 'irreg':
            if self._dim != 2:
                raise RuntimeError('only 2D irregular binned likelihood is supported!')
            self._bins[0] = jnp.array(self._bins[0])
            self._bins[1] = jnp.array(self._bins[1])
            self.component_bins_type = 'irreg'
            # x-binning should 1 longer than y-binning
            mask0 = len(self._bins[0]) != len(self._bins[1]) + 1
            # all y-binning should have the same length
            mask1 = not all(len(b) == len(self._bins[1][0]) for b in self._bins[1])
            if mask0 or mask1:
                raise ValueError(f'Please check the binning in {self.name}!')
            self.data_hist = make_hist_irreg_bin_2d(
                self.data,
                bins_x=self._bins[0],
                bins_y=self._bins[1],
                weights=jnp.ones(len(self.data)),
            )
        else:
            raise ValueError("'bins_type' should either be meshgrid, equiprob or irreg")

    def __getitem__(self, keys):
        """Get component in likelihood"""
        return self.components[keys]

    def register_component(self,
                           component_cls: Component,
                           component_name: str,
                           file_name: str = None):
        """Create an appletree likelihood.

        :param component_cls: class of Component
        :param component_name: name of Component
        :param file_name: file used in ComponentFixed
        """
        if component_name in self.components:
            raise ValueError(f'Component named {component_name} already existed!')

        # Initialize component
        component = component_cls(
            name=component_name,
            llh_name=self.name,
            bins=self._bins,
            bins_type=self.component_bins_type,
            file_name=file_name,
        )
        component.rate_name = component_name + '_rate'
        kwargs = {'data_names': self._bins_on}
        if isinstance(component, ComponentSim):
            kwargs['func_name'] = self.name + '_' + component_name + '_sim'
            kwargs['data_names'] = self._bins_on + ['eff']
        component.deduce(**kwargs)
        component.compile()

        # Update components sheet
        self.components[component_name] = component

        # Update needed parameters
        self.needed_parameters |= self.components[component_name].needed_parameters

    def get_log_likelihood(self, key, batch_size, parameters):
        """Get log likelihood of given parameters.

        :param key: a pseudo-random number generator (PRNG) key
        :param batch_size: int of number of simulated events
        :param parameters: dict of parameters used in simulation
        """
        key, model_hist = self._simulate_model_hist(key, batch_size, parameters)
        # Poisson likelihood
        llh = jnp.sum(self.data_hist * jnp.log(model_hist) - model_hist)
        llh = float(llh)
        if np.isnan(llh):
            llh = -np.inf
        return key, llh

    def print_likelihood_summary(self,
                                 indent: str = ' '*4,
                                 short: bool = True):
        """Print likelihood summary: components, bins, file names.

        :param indent: str of indent
        :param short: bool, whether only print short summary
        """
        print('\n'+'-'*40)

        print(f'BINNING\n')
        print(f'{indent}bins_type: {self._bins_type}')
        print(f'{indent}bins_on: {self._bins_on}')
        if not short:
            print(f'{indent}bins: {self._bins}')
        print('\n'+'-'*40)

        print(f'DATA\n')
        print(f'{indent}file_name: {self._data_file_name}')
        print(f'{indent}data_rate: {float(self.data_hist.sum())}')
        print('\n'+'-'*40)

        print('MODEL\n')
        for i, component_name in enumerate(self.components):
            name = component_name
            component = self[component_name]
            need = component.needed_parameters

            print(f'{indent}COMPONENT {i}: {name}')
            if isinstance(component, ComponentSim):
                print(f'{indent*2}type: simulation')
                print(f'{indent*2}rate_par: {component.rate_name}')
                print(f'{indent*2}pars: {need}')
                if not short:
                    print(f'{indent*2}worksheet: {component.worksheet}')
            elif isinstance(component, ComponentFixed):
                print(f'{indent*2}type: fixed')
                print(f'{indent*2}file_name: {component._file_name}')
                print(f'{indent*2}rate_par: {component.rate_name}')
                print(f'{indent*2}pars: {need}')
                if not short:
                    print(f'{indent*2}from_file: {component.file_name}')
            else:
                pass
            print()

        print('-'*40)

    def _sanity_check(self):
        """Check equality between number of bins group and observables"""
        if len(self._bins_on) != len(self._bins):
            raise RuntimeError('Length of bins must be the same as length of bins_on!')

    def _simulate_model_hist(self, key, batch_size, parameters):
        """Histogram of simulated observables.

        :param key: a pseudo-random number generator (PRNG) key
        :param batch_size: int of number of simulated events
        :param parameters: dict of parameters used in simulation
        """
        hist = jnp.zeros_like(self.data_hist)
        for component_name, component in self.components.items():
            if isinstance(component, ComponentSim):
                key, _hist = component.simulate_hist(key, batch_size, parameters)
            elif isinstance(component, ComponentFixed):
                _hist = component.simulate_hist(parameters)
            else:
                raise TypeError(f'unsupported component type for {component_name}!')
            hist += _hist
        return key, hist

    def simulate_weighed_data(self, key, batch_size, parameters):
        result = []
        for component_name, component in self.components.items():
            if isinstance(component, ComponentSim):
                key, _result = component.simulate_weighed_data(key, batch_size, parameters)
            elif isinstance(component, ComponentFixed):
                _result = component.simulate_weighed_data(parameters)
            else:
                raise TypeError(f'unsupported component type for {component_name}!')
            result.append(_result)
        result = [r for r in np.hstack(result)]
        return key, result



import os
from functools import partial
from time import time

import jax
from jax import numpy as jnp
import numpy as np
from jax import random, lax, jit, vmap
from numpyro.distributions.util import _binomial_dispatch as _binomial_dispatch_numpyro

from appletree.utils import exporter

export, __all__ = exporter(export_self=False)

INT = np.int32
FLOAT = np.float32
ALWAYS_USE_NORMAL_APPROX_IN_BINOM = bool(
    os.environ.get('ALWAYS_USE_NORMAL_APPROX_IN_BINOM', True))


@export
def get_key(seed=None):
    """Generate a key for jax.random."""
    if seed is None:
        seed = int(time()*1e6)
    return random.PRNGKey(seed)


@export
@partial(jit, static_argnums=(3, ))
def uniform(key, vmin, vmax, shape=()):
    """Uniform random sampler.

    :param key: seed for random generator.
    :param vmin: <jnp.array>-like min in uniform distribution.
    :param vmax: <jnp.array>-like max in uniform distribution.
    :param shape: output shape. If not given, output has shape
        jnp.broadcast_shapes(jnp.shape(vmin), jnp.shape(vmax)).
    :return: an updated seed, random variables.
    """
    key, seed = random.split(key)

    shape = shape or jnp.broadcast_shapes(jnp.shape(vmin), jnp.shape(vmax))
    vmin = jnp.broadcast_to(vmin, shape).astype(FLOAT)
    vmax = jnp.broadcast_to(vmax, shape).astype(FLOAT)

    rvs = random.uniform(seed, shape, minval=vmin, maxval=vmax)
    return key, rvs.astype(FLOAT)


@export
@partial(jit, static_argnums=(2, ))
def poisson(key, lam, shape=()):
    """Poisson random sampler.

    :param key: seed for random generator.
    :param lam: <jnp.array>-like expectation in poisson distribution.
    :param shape: output shape. If not given, output has shape jnp.shape(lam).
    :return: an updated seed, random variables.
    """
    key, seed = random.split(key)

    shape = shape or jnp.shape(lam)
    lam = jnp.broadcast_to(lam, shape).astype(FLOAT)

    rvs = random.poisson(seed, lam, shape=shape)
    return key, rvs.astype(INT)


@export
@partial(jit, static_argnums=(3, ))
def normal(key, mean, std, shape=()):
    """Normal distribution random sampler.

    :param key: seed for random generator.
    :param mean: <jnp.array>-like mean in normal distribution.
    :param std: <jnp.array>-like std in normal distribution.
    :param shape: output shape. If not given, output has shape
    jnp.broadcast_shapes(jnp.shape(mean), jnp.shape(std)).
    :return: an updated seed, random variables.
    """
    key, seed = random.split(key)

    shape = shape or jnp.broadcast_shapes(jnp.shape(mean), jnp.shape(std))
    mean = jnp.broadcast_to(mean, shape).astype(FLOAT)
    std = jnp.broadcast_to(std, shape).astype(FLOAT)

    rvs = random.normal(seed, shape=shape) * std + mean
    return key, rvs.astype(FLOAT)


@export
@partial(jit, static_argnums=(5, ))
def truncate_normal(key, mean, std, vmin=None, vmax=None, shape=()):
    """Truncated normal distribution random sampler.

    :param key: seed for random generator.
    :param mean: <jnp.array>-like mean in normal distribution.
    :param std: <jnp.array>-like std in normal distribution.
    :param vmin: <jnp.array>-like min value to clip.
        By default it's None. vmin and vmax cannot be both None.
    :param vmax: <jnp.array>-like max value to clip.
        By default it's None. vmin and vmax cannot be both None.
    :param shape: parameter passed to normal(..., shape=shape)
    :return: an updated seed, random variables.
    """
    key, rvs = normal(key, mean, std, shape=shape)
    rvs = jnp.clip(rvs, a_min=vmin, a_max=vmax)
    return key, rvs.astype(FLOAT)


@export
@partial(jit, static_argnums=(2, ))
def bernoulli(key, p, shape=()):
    """Bernoulli random sampler.
    :param key: seed for random generator.
    :param p: <jnp.array>-like probability in bernoulli distribution.
    :param shape: output shape. If not given, output has shape jnp.shape(lam).
    :return: an updated seed, random variables.
    """
    key, seed = random.split(key)

    shape = shape or jnp.shape(p)
    p = jnp.broadcast_to(p, shape).astype(FLOAT)

    rvs = random.bernoulli(seed, p, shape=shape)
    return key, rvs.astype(INT)


@export
@partial(jit, static_argnums=(3, 4))
def binomial(key, p, n, shape=(), always_use_normal=ALWAYS_USE_NORMAL_APPROX_IN_BINOM):
    """Binomial random sampler.

    :param key: seed for random generator.
    :param p: <jnp.array>-like probability in binomial distribution.
    :param n: <jnp.array>-like count in binomial distribution.
    :param shape: output shape. If not given, output has shape
        jnp.broadcast_shapes(jnp.shape(p), jnp.shape(n)).
    :param always_use_normal: If true, then Norm(np, sqrt(npq)) is always used.
        Otherwise if np < 5, use the inversion method instead.
    :return: an updated seed, random variables.
    """
    def _binomial_normal_approx_dispatch(seed, p, n):
        q = 1. - p
        mean = n * p
        std = jnp.sqrt(n * p * q)
        rvs = jnp.clip(random.normal(seed) * std + mean, a_min=0.)
        return rvs.round().astype(INT)

    def _binomial_dispatch(seed, p, n):
        use_normal_approx = (n * p >= 5.)
        return lax.cond(
            use_normal_approx,
            (seed, p, n), lambda x: _binomial_normal_approx_dispatch(*x),
            (seed, p, n), lambda x: _binomial_dispatch_numpyro(*x),
        )

    key, seed = random.split(key)

    shape = shape or lax.broadcast_shapes(jnp.shape(p), jnp.shape(n))
    p = jnp.reshape(jnp.broadcast_to(p, shape), -1).astype(FLOAT)
    n = jnp.reshape(jnp.broadcast_to(n, shape), -1).astype(INT)
    seed = random.split(seed, jnp.size(p))

    if always_use_normal:
        dispatch = _binomial_normal_approx_dispatch
    else:
        dispatch = _binomial_dispatch

    if jax.default_backend() == "cpu":
        ret = lax.map(lambda x: dispatch(*x), (seed, p, n))
    else:
        ret = vmap(lambda *x: dispatch(*x))(seed, p, n)
    return key, jnp.reshape(ret, shape).astype(INT)


@export
@jit
def uniform_key_vectorized(key):
    """Uniform(0,1) distribution sampler, vectorized by key.
    Note: key won't be updated!

    :param key: seed for random generator, with shape (N, 2)
    :return: random varibles with shape (N, )
    """
    sampler = vmap(jax.random.uniform, (0, ), 0)
    return sampler(key)


class TwoHalfNorm:
    """Continuous distribution, two half Normal"""

    @staticmethod
    def rvs(mu=0, sigma_pos=1, sigma_neg=1, size=1):
        """
        Get random variables
        :param mu: float, 'center' value of the distribution
        :param sigma_pos: float,
        Standard deviation of the distribution when variable larger than mu. Must be non-negative.
        :param sigma_neg: float,
        Standard deviation of the distribution when variable smaller than mu. Must be non-negative.
        :param size: int or tuple of ints, Output shape.
        :return: random samples
        """
        assert (sigma_pos > 0) and (sigma_neg > 0), 'sigma should be positive'
        pos_half_prob = sigma_pos / (sigma_pos + sigma_neg)

        use_pos_half = np.random.uniform(size=size) < pos_half_prob
        use_neg_half = ~use_pos_half

        n_sigma = np.abs(np.random.normal(size=size))
        offset = use_pos_half * n_sigma * sigma_pos - use_neg_half * n_sigma * sigma_neg

        return offset + mu

    @staticmethod
    def logpdf(x, mu=0, sigma_pos=1, sigma_neg=1):
        """
        Log of the probability density function.
        :param x: array, input variables
        :param mu: float, 'center' value of the distribution
        :param sigma_pos: float,
        Standard deviation of the distribution when variable larger than mu. Must be non-negative.
        :param sigma_neg: float,
        Standard deviation of the distribution when variable smaller than mu. Must be non-negative.
        :param size: int or tuple of ints, Output shape.
        :return: log probability density function
        """
        assert (sigma_pos > 0) and (sigma_neg > 0), 'sigma should be positive'
        norm = 2 / (sigma_pos + sigma_neg) / np.sqrt(2 * np.pi)
        logpdf = np.where(x < mu, -(x - mu)**2 / sigma_neg**2 / 2, -(x - mu)**2 / sigma_pos**2 / 2)
        logpdf += np.log(norm)
        return logpdf



import typing as ty

from immutabledict import immutabledict
from jax import numpy as jnp
from warnings import warn

from appletree.share import _cached_configs
from appletree.utils import exporter, load_json, get_file_path, integrate_midpoint, cum_integrate_midpoint
from appletree import interpolation
from appletree.interpolation import FLOAT_POS_MIN, FLOAT_POS_MAX

export, __all__ = exporter()

OMITTED = '<OMITTED>'

__all__ += 'OMITTED'.split()


@export
def takes_config(*configs):
    """Decorator for plugin classes, to specify which configs it takes.

    :param configs: Config instances of configs this plugin takes.
    """

    def wrapped(plugin_class):
        """
        :param plugin_class: plugin needs configuration
        """
        result = {}
        for config in configs:
            if not isinstance(config, Config):
                raise RuntimeError("Specify config options by Config objects")
            config.taken_by = plugin_class.__name__
            result[config.name] = config

        if (hasattr(plugin_class, 'takes_config') and len(plugin_class.takes_config)):
            # Already have some configs set, e.g. because of subclassing
            # where both child and parent have a takes_config decorator
            for config in result.values():
                if config.name in plugin_class.takes_config:
                    raise RuntimeError(
                        f'Attempt to specify config {config.name} twice')
            plugin_class.takes_config = immutabledict({
                **plugin_class.takes_config, **result})
        else:
            plugin_class.takes_config = immutabledict(result)

        # Should set the configurations as the attributes of Plugin
        return plugin_class

    return wrapped


@export
class Config():
    """Configuration option taken by a appletree plugin"""

    def __init__(self,
                 name: str,
                 type: ty.Union[type, tuple, list] = OMITTED,
                 default: ty.Any = OMITTED,
                 help: str = ''):
        """Initialization.

        :param name: name of the map
        :param type: Excepted type of the option's value.
        :param default: Default value the option takes.
        :param help: description of the map
        """
        self.name = name
        self.type = type
        self.default = default
        self.help = help

        # Sanity check
        if isinstance(self.default, dict):
            raise ValueError(
                f"Do not set {self.name}'s default value as dict!",
            )

    def get_default(self):
        """Get default value of configuration"""
        if self.default is not OMITTED:
            return self.default

        raise ValueError(f"Missing option {self.name} "
                         f"required by {self.taken_by}")

    def build(self, llh_name: str = None):
        """Build configuration, set attributes to Config instance"""
        raise NotImplementedError


@export
class Constant(Config):
    """Constant is a special config which takes only certain value"""

    value = None

    def build(self, llh_name: str = None):
        """Set value of Constant"""
        if self.name in _cached_configs:
            value = _cached_configs[self.name]
        else:
            value = self.get_default()
            # Update values to sharing dictionary
            _cached_configs[self.name] = value

        if isinstance(value, dict):
            try:
                self.value = value[llh_name]
            except KeyError:
                mesg = f'You specified {self.name} as a dictionary. '
                mesg += f'The key of it should be the name of one '
                mesg += f'of the likelihood, '
                mesg += f'but it is {llh_name}.'
                raise ValueError(mesg)
        else:
            self.value = value


@export
class Map(Config):
    """
    Map is a special config that takes input files.
    The method `apply` is dynamically assigned.
    When using points, the `apply` will be `map_point`, 
    while using regular binning, the `apply` will be `map_regbin`.
    When using log-binning, we will first convert the positions to log space.
    """

    def build(self, llh_name: str = None):
        """Cache the map to jnp.array"""

        if self.name in _cached_configs:
            file_path = _cached_configs[self.name]
        else:
            file_path = get_file_path(self.get_default())
            # Update values to sharing dictionary
            _cached_configs[self.name] = file_path

        if isinstance(file_path, dict):
            try:
                self.file_path = file_path[llh_name]
            except KeyError:
                mesg = f'You specified {self.name} as a dictionary. '
                mesg += f'The key of it should be the name of one '
                mesg += f'of the likelihood, '
                mesg += f'but it is {llh_name}.'
                raise ValueError(mesg)
        else:
            self.file_path = file_path

        data = load_json(self.file_path)

        coordinate_type = data['coordinate_type']
        if coordinate_type == 'point' or coordinate_type == 'log_point':
            self.build_point(data)
        elif coordinate_type == 'regbin' or coordinate_type == 'log_regbin':
            self.build_regbin(data)
        else:
            raise ValueError("map_type must be either 'point' or 'regbin'!")

    def build_point(self, data):
        """Cache the map to jnp.array if bins_type is point"""
        if data['coordinate_name'] == 'pdf' or data['coordinate_name'] == 'cdf':
            if data['coordinate_type'] == 'log_point':
                raise ValueError(
                    f'It is not a good idea to use log pdf nor cdf '
                    f'in map {self.file_path}. '
                    f'Because its coordinate type is log-binned. '
                )

        if data['coordinate_name'] == 'pdf':
            warn(f'Convert {self.name} from (x, pdf) to (cdf, x).')
            x, cdf = self.pdf_to_cdf(data['coordinate_system'], data['map'])
            data['coordinate_name'] = 'cdf'
            data['coordinate_system'] = cdf
            data['map'] = x

        self.coordinate_type = data['coordinate_type']
        self.coordinate_name = data['coordinate_name']
        self.coordinate_system = jnp.asarray(data['coordinate_system'], dtype=float)
        self.map = jnp.asarray(data['map'], dtype=float)

        setattr(self, 'interpolator', interpolation.curve_interpolator)
        if self.coordinate_type == 'log_point':
            if jnp.any(self.coordinate_system <= 0):
                raise ValueError(
                    f'Find non-positive coordinate system in map {self.file_path}, '
                    f'which is specified as {self.coordinate_type}'
                )
            setattr(self, 'preprocess', self.log_pos)
        else:
            setattr(self, 'preprocess', self.linear_pos)
        setattr(self, 'apply', self.map_point)

    def map_point(self, pos):
        val = self.interpolator(
            self.preprocess(pos),
            self.preprocess(self.coordinate_system),
            self.preprocess(self.map),
        )
        return val

    def build_regbin(self, data):
        """Cache the map to jnp.array if bins_type is regbin"""
        if 'pdf' in data['coordinate_name'] or 'cdf' in data['coordinate_name']:
            if data['coordinate_type'] == 'log_regbin':
                raise ValueError(
                    f'It is not a good idea to use log pdf nor cdf '
                    f'in map {self.file_path}. '
                    f'Because its coordinate type is log-binned. '
                )

        self.coordinate_type = data['coordinate_type']
        self.coordinate_name = data['coordinate_name']
        self.coordinate_lowers = jnp.asarray(data['coordinate_lowers'], dtype=float)
        self.coordinate_uppers = jnp.asarray(data['coordinate_uppers'], dtype=float)
        self.map = jnp.asarray(data['map'], dtype=float)

        if len(self.coordinate_lowers) == 1:
            setattr(self, 'interpolator', interpolation.map_interpolator_regular_binning_1d)
        elif len(self.coordinate_lowers) == 2:
            setattr(self, 'interpolator', interpolation.map_interpolator_regular_binning_2d)
        elif len(self.coordinate_lowers) == 3:
            setattr(self, 'interpolator', interpolation.map_interpolator_regular_binning_3d)
        if self.coordinate_type == 'log_regbin':
            if jnp.any(self.coordinate_lowers <= 0) or jnp.any(self.coordinate_uppers <= 0):
                raise ValueError(
                    f'Find non-positive coordinate system in map {self.file_path}, '
                    f'which is specified as {self.coordinate_type}'
                )
            setattr(self, 'preprocess', self.log_pos)
        else:
            setattr(self, 'preprocess', self.linear_pos)
        setattr(self, 'apply', self.map_regbin)

    def map_regbin(self, pos):
        val = self.interpolator(
            self.preprocess(pos),
            self.preprocess(self.coordinate_lowers),
            self.preprocess(self.coordinate_uppers),
            self.map,
        )
        return val

    def linear_pos(self, pos):
        return pos

    def log_pos(self, pos):
        return jnp.log10(jnp.clip(pos, a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX))

    def pdf_to_cdf(self, x, pdf):
        """Convert pdf map to cdf map"""
        norm = integrate_midpoint(x, pdf)
        x, cdf = cum_integrate_midpoint(x, pdf)
        cdf /= norm
        return x, cdf


@export
class SigmaMap(Config):
    """
    Maps with uncertainty.
    Default value is a list whose order is:
    [median, lower, upper, (parameter)]
    Each map is assigned as attribute of SigmaMap.
    If the last element in the list is the required parameter.
    """

    def build(self, llh_name: str = None):
        """Read maps"""
        self.llh_name = llh_name
        if self.name in _cached_configs:
            _configs = _cached_configs[self.name]
        else:
            _configs = self.get_default()
            # Update values to sharing dictionary
            _cached_configs[self.name] = _configs

        if isinstance(_configs, dict):
            try:
                self._configs = _configs[llh_name]
            except KeyError:
                mesg = f'You specified {self.name} as a dictionary. '
                mesg += f'The key of it should be the name of one '
                mesg += f'of the likelihood, '
                mesg += f'but it is {llh_name}.'
                raise ValueError(mesg)
        else:
            self._configs = _configs

        self._configs_default = self.get_default()

        maps = {}
        sigmas = ['median', 'lower', 'upper']
        for i, sigma in enumerate(sigmas):
            maps[sigma] = Map(
                name=self.name + f'_{sigma}',
                default=self._configs_default[i])
            if maps[sigma].name not in _cached_configs.keys():
                _cached_configs[maps[sigma].name] = {}
            if isinstance(_cached_configs[maps[sigma].name], dict):
                # In case some plugins only use the median
                # and may already update the map name in `_cached_configs`
                _cached_configs[maps[sigma].name].update(
                    {self.llh_name: self._configs[i]})
            setattr(self, sigma, maps[sigma])

        self.median.build(llh_name=self.llh_name)
        self.lower.build(llh_name=self.llh_name)
        self.upper.build(llh_name=self.llh_name)

        if len(self._configs) > 4:
            raise ValueError(f'You give too much information in {self.name} configs.')

        # Find required parameter
        if len(self._configs) == 4:
            self.required_parameter = self._configs[-1]
            print(
                f'{self.llh_name} is using the parameter '
                f'{self.required_parameter} in {self.name} map.')
        else:
            self.required_parameter = self.name + '_sigma'

    def apply(self, pos, parameters):
        """Apply SigmaMap with sigma and position"""
        sigma = parameters[self.required_parameter]
        median = self.median.apply(pos)
        lower = self.lower.apply(pos)
        upper = self.upper.apply(pos)
        add_pos = (upper - median) * sigma
        add_neg = (median - lower) * sigma
        add = jnp.where(sigma > 0, add_pos, add_neg)
        return median + add



import inspect
from copy import deepcopy

from immutabledict import immutabledict

from appletree import utils
from appletree.utils import exporter

export, __all__ = exporter()


@export
class Plugin():
    """The smallest simulation unit."""

    # Do not initialize this class because it is base
    __is_base = True

    # the plugin's dependency(the arguments of `simulate`)
    depends_on = []

    # the plugin can provide(`simulate` will return)
    provides = []

    # relevant parameters, will be fitted in MCMC
    parameters = ()

    # Set using the takes_config decorator
    takes_config = immutabledict()

    def __init__(self, llh_name: str = None):
        """Initialization."""
        # llh_name will tell us which map to use
        self.llh_name = llh_name
        if not self.depends_on:
            raise ValueError('depends_on not provided for '
                             f'{self.__class__.__name__}')

        if not self.provides:
            raise ValueError('provides not provided for '
                             f'{self.__class__.__name__}')

        # configs are loaded when a plugin is initialized
        for config in self.takes_config.values():
            config.build(self.llh_name)

        self.sanity_check()

        # Do not set configurations as static
        for config in self.takes_config.values():
            setattr(self, config.name, deepcopy(config))

    def __call__(self, *args, **kwargs):
        """Calls self.simulate"""
        return self.simulate(*args, **kwargs)

    def simulate(self, *args, **kwargs):
        """The main simulation function.

        :param key: a jnp.array with length 2, used to generate random variables.
            See https://jax.readthedocs.io/en/latest/jax-101/05-random-numbers.html
        :param parameters: a dictionary with key being parameters' names. Plugin will
            get values of self.parameters from this dictionary.
        :param args: other args following `key` and `parameters` must be in the order of
            self.depends_on.
        :return: `key` and output simulated variables, ordered by self.provides. `key` will
            be updated if it's used inside self.simulate to generate random variables.
        """
        raise NotImplementedError

    def sanity_check(self):
        """Check the consistency between `depends_on`, `provides` and in(out)put of `self.simulate`"""
        arguments = inspect.getfullargspec(self.simulate)[0]
        if arguments[1] != 'key':
            mesg = f"First argument of {self.__class__.__name__}"
            mesg += ".simulate should be 'key'."
            raise ValueError(mesg)
        if arguments[2] != 'parameters':
            mesg = f"Second argument of {self.__class__.__name__}"
            mesg += ".simulate should be 'parameters'."
            raise ValueError(mesg)
        for i, depend in enumerate(self.depends_on, start=3):
            if arguments[i] != depend:
                mesg = f"{i}th argument of {self.__class__.__name__}"
                mesg += f".simulate should be '{depend}'."
                mesg += f'Plugin {self.__class__.__name__} is insane, check dependency!'
                raise ValueError(mesg)



@export
def add_plugin_extensions(module1, module2, force=False):
    """Add plugins of module2 to module1"""
    utils.add_extensions(module1, module2, Plugin, force=force)


@export
def _add_plugin_extension(module, plugin, force=False):
    """Add plugin to module"""
    utils._add_extension(module, plugin, Plugin, force=force)



from jax import numpy as jnp
from jax import jit
from functools import partial

from appletree import randgen
from appletree.plugin import Plugin
from appletree.utils import exporter

export, __all__ = exporter(export_self=False)


@export
class Quanta(Plugin):
    depends_on = ['energy']
    provides = ['num_quanta']
    parameters = ('w', 'fano',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy):
        num_quanta_mean = energy / parameters['w']
        num_quanta_std = jnp.sqrt(num_quanta_mean * parameters['fano'])
        key, num_quanta = randgen.truncate_normal(key, num_quanta_mean, num_quanta_std, vmin=0)
        return key, num_quanta.round().astype(int)


@export
class IonizationER(Plugin):
    depends_on = ['num_quanta']
    provides = ['num_ion']
    parameters = ('nex_ni_ratio',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_quanta):
        p_ion = 1. / (1. + parameters['nex_ni_ratio'])
        key, num_ion = randgen.binomial(key, p_ion, num_quanta)
        return key, num_ion


@export
class mTI(Plugin):
    depends_on = ['energy']
    provides = ['recomb_mean']
    parameters = ('w', 'nex_ni_ratio', 'py0', 'py1', 'py2', 'py3', 'py4', 'field')

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy):
        ni = energy / parameters['w'] / (1. + parameters['nex_ni_ratio'])
        ti = ni * parameters['py0'] * jnp.exp(- energy / parameters['py1']) * parameters['field']**parameters['py2'] / 4.
        fd = 1. / (1. + jnp.exp(- (energy - parameters['py3']) / parameters['py4']))
        r = jnp.where(
            ti < 1e-2,
            ti / 2. - ti * ti / 3.,
            1. - jnp.log(1. + ti) / ti,
        )
        return key, r * fd


@export
class RecombFluct(Plugin):
    depends_on = ['energy']
    provides = ['recomb_std']
    parameters = ('rf0', 'rf1')

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy):
        fd_factor = 1. - jnp.exp(- energy / parameters['rf1'])
        recomb_std = jnp.clip(parameters['rf0'] * fd_factor, 0, 1.)
        return key, recomb_std


@export
class TrueRecombER(Plugin):
    depends_on = ['recomb_mean', 'recomb_std']
    provides = ['recomb']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, recomb_mean, recomb_std):
        key, recomb = randgen.truncate_normal(key, recomb_mean, recomb_std, vmin=0., vmax=1.)
        return key, recomb


@export
class RecombinationER(Plugin):
    depends_on = ['num_quanta', 'num_ion', 'recomb']
    provides = ['num_photon', 'num_electron']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_quanta, num_ion, recomb):
        p_not_recomb = 1. - recomb
        key, num_electron = randgen.binomial(key, p_not_recomb, num_ion)
        num_photon = num_quanta - num_electron
        return key, num_photon, num_electron



import copy
import json

import numpy as np

from appletree.randgen import TwoHalfNorm


class Parameter():
    """Parameter handler to update parameters and calculate prior."""

    def __init__(self, parameter_config):
        """Initialization

        :param parameter_config: can be either

          * str: the json file name where the config is stored.
          * dict: config dictionary.
        """
        if isinstance(parameter_config, str):
            with open(parameter_config, 'r') as file:
                self.par_config = json.load(file)
        elif isinstance(parameter_config, dict):
            self.par_config = copy.deepcopy(parameter_config)
        else:
            raise RuntimeError('Parameter configuration should be file name or dictionary')

        self._parameter_fixed = set()
        self._parameter_fit = set()
        self.init_parameter()

    def init_parameter(self, seed=None):
        """Initializing parameters by sampling prior.
        If the prior is free, then sampling from the initial guess.

        :param seed: integer, sent to np.random.seed(seed)
        """
        self._parameter_dict = {par_name: 0 for par_name in self.par_config.keys()}

        for par_name in self.par_config.keys():
            if self.par_config[par_name]['prior_type'] == 'fixed':
                self._parameter_fixed.add(par_name)
            else:
                self._parameter_fit.add(par_name)
        # Parameters name set is not sorted
        self._parameter_fixed = set(self._parameter_fixed)
        self._parameter_fit = set(self._parameter_fit)

        if seed is not None:
            np.random.seed(seed)
        self.sample_prior()

    @property
    def parameter_fit(self):
        """Return sorted list of parameters name waiting for fitting"""
        return sorted(self._parameter_fit)

    def sample_prior(self):
        """Sampling parameters from prior and set self._parameter_dict.
        If the prior is free, then sampling from the initial guess.
        """
        for par_name in self._parameter_dict:
            try:
                setting = self.par_config[par_name]
            except KeyError:
                raise RuntimeError(f'Requested parameter "{par_name}" not in given configuration')

            args = setting['prior_args']
            prior_type = setting['prior_type']

            if prior_type == 'norm':
                kwargs = {
                    'loc': args['mean'],
                    'scale': args['std'],
                }
                val = np.random.normal(**kwargs)
                self._parameter_dict[par_name] = np.clip(val, *setting['allowed_range'])
            elif prior_type == 'twohalfnorm':
                kwargs = {
                    'mu': args['mu'],
                    'sigma_pos': args['sigma_pos'],
                    'sigma_neg': args['sigma_neg'],
                }
                val = TwoHalfNorm.rvs(**kwargs)
                self._parameter_dict[par_name] = np.clip(val, *setting['allowed_range'])
            elif prior_type == 'uniform':
                kwargs = {
                    'low': args['lower'],
                    'high': args['upper'],
                }
                val = np.random.uniform(**kwargs)
                self._parameter_dict[par_name] = np.clip(val, *setting['allowed_range'])
            elif prior_type == 'free':
                kwargs = {
                    'loc': setting['init_mean'],
                    'scale': setting['init_std'],
                }
                val = np.random.normal(**kwargs)
                self._parameter_dict[par_name] = np.clip(val, *setting['allowed_range'])
            elif prior_type == 'fixed':
                self._parameter_dict[par_name] = args['val']

    def sample_init(self):
        """Samping parameters from initial guess clipped
        by the allowed_range and set self._parameter_dict.
        """
        for par_name in self._parameter_dict:
            try:
                setting = self.par_config[par_name]
            except KeyError:
                raise RuntimeError(f'Requested parameter "{par_name}" not in given configuration')

            args = setting['prior_args']
            prior_type = setting['prior_type']

            if prior_type == 'fixed':
                self._parameter_dict[par_name] = args['val']
            else:
                kwargs = {
                    'loc': setting['init_mean'],
                    'scale': setting['init_std'],
                }
                val = np.random.normal(**kwargs)
                self._parameter_dict[par_name] = np.clip(val, *setting['allowed_range'])

    @property
    def log_prior(self):
        """Return log prior. If any parameter is out of allowed_range return -np.inf."""
        log_prior = 0

        for par_name in self._parameter_fit:
            val = self._parameter_dict[par_name]
            setting = self.par_config[par_name]

            args = setting['prior_args']
            prior_type = setting['prior_type']

            if val < setting['allowed_range'][0] or val > setting['allowed_range'][1]:
                log_prior += -np.inf
            elif prior_type == 'norm':
                mean = args['mean']
                std = args['std']
                log_prior += - (val - mean)**2 / 2 / std**2
            elif prior_type == 'twohalfnorm':
                mu = args['mu']
                sigma_pos = args['sigma_pos']
                sigma_neg = args['sigma_neg']
                log_prior += TwoHalfNorm.logpdf(
                    x=val,
                    mu=mu,
                    sigma_pos=sigma_pos,
                    sigma_neg=sigma_neg,
                )
            elif prior_type == 'free':
                pass
            elif prior_type == 'uniform':
                pass

        return log_prior

    def check_parameter_exist(self, keys, return_not_exist=False):
        """Check whether the keys exist in parameters.

        :param keys: Parameter names. Can be a single str, or a list of str.
        :param return_not_exist: If False, function will return a bool if all keys exist.
            If True, function will additionally return the list of the not existing keys.
        """
        if isinstance(keys, (set, list)):
            not_exist = []
            for key in keys:
                if key not in self._parameter_dict:
                    not_exist.append(key)
            all_exist = (not_exist == [])
            if return_not_exist:
                return (all_exist, not_exist)
            else:
                return (all_exist)
        elif isinstance(keys, str):
            if return_not_exist:
                return (keys in self._parameter_dict, keys)
            else:
                return (keys in self._parameter_dict)
        elif isinstance(keys, dict):
            return self.check_parameter_exist(list(keys.keys()), return_not_exist)

        else:
            raise ValueError("keys must be a str or a list of str!")

    def set_parameter(self, keys, vals=None):
        """Set parameter values.

        :param keys: Parameter names. Can be either

          * str: vals must be int or float.
          * list: vals must have the same length.
          * dict: vals will be overwritten as keys.values().
        :param vals: Values to be set.
        """
        all_exist, not_exist = self.check_parameter_exist(keys, return_not_exist=True)
        assert all_exist, f"{not_exist} not found!"

        if isinstance(keys, list):
            assert len(keys) == len(vals), "keys must have the same length as vals!"
            for key, val in zip(keys, vals):
                self._parameter_dict[key] = val
        elif isinstance(keys, dict):
            self.set_parameter(list(keys.keys()), keys.values())
        elif isinstance(keys, str):
            assert isinstance(vals, (float, int)), "val must be either float or int!"
            self._parameter_dict[keys] = vals
        else:
            raise ValueError("keys must be a str or a list of str!")

    def get_parameter(self, keys):
        """Return parameter values.

        :param keys: Parameter names. Can be a single str, or a list of str.
        """
        all_exist, not_exist = self.check_parameter_exist(keys, return_not_exist=True)
        assert all_exist, f"{not_exist} not found!"

        return self.__getitem__(keys)

    def __getitem__(self, keys):
        """__getitem__, keys can be str/list/set"""
        if isinstance(keys, (set, list)):
            return np.array([self._parameter_dict[key] for key in keys])
        elif isinstance(keys, str):
            return self._parameter_dict[keys]
        else:
            raise ValueError("keys must be a str or a list of str!")

    @property
    def parameter_fit_array(self):
        """Return non-fixed parameters, ordered by self._parameter_fit."""
        return self.get_parameter(self.parameter_fit)

    def get_all_parameter(self):
        """Return all parameters as a dict."""
        return self._parameter_dict



from appletree.context import Context
from appletree.utils import get_file_path


class ContextRn220(Context):
    """A specified context for ER response by Rn220 fit"""

    def __init__(self):
        """Initialization"""
        config = get_file_path('rn220_sr0.json')
        super().__init__(config)


class ContextRn220Ar37(Context):
    """A specified context for ER response by Rn220 & Ar37 combined fit"""

    def __init__(self):
        """Initialization"""
        config = get_file_path('rn220_ar37_sr0.json')
        super().__init__(config)



from . import common
from .common import *

from . import er_microphys
from .er_microphys import *

from . import lyqy
from .lyqy import *

from . import detector
from .detector import *

from . import reconstruction
from .reconstruction import *

from . import efficiency
from .efficiency import *



from jax import numpy as jnp
from jax import jit, vmap

from appletree.utils import exporter

export, __all__ = exporter(export_self=False)


@export
@jit
def make_hist_mesh_grid(sample, bins=10, weights=None):
    """Same as jnp.histogramdd."""
    hist, _ = jnp.histogramdd(sample, bins=bins, weights=weights)
    return hist


@export
@jit
def make_hist_irreg_bin_2d(sample, bins_x, bins_y, weights):
    """Make a histogram with irregular binning.

    :param sample: array with shape (N, 2)
    :param bins_x: array with shape (M1, )
    :param bins_y: array with shape (M1-1, M2)
    :param weights: array with shape (N, )
    """
    x = sample[:, 0]
    y = sample[:, 1]

    ind_x = jnp.searchsorted(bins_x, x)
    ind_y = vmap(jnp.searchsorted, (0, 0), 0)(bins_y[ind_x-1], y)

    bin_ind = jnp.stack((ind_x, ind_y))

    output_shape = (len(bins_x)+1, bins_y.shape[-1]+1)
    hist = jnp.zeros(output_shape)
    hist = hist.at[tuple(bin_ind)].add(weights)

    return hist[1:-1, 1:-1]



from jax import numpy as jnp
from jax import jit
from functools import partial

import appletree
from appletree.plugin import Plugin
from appletree.config import SigmaMap
from appletree.utils import exporter

export, __all__ = exporter(export_self=False)


@export
class S2Threshold(Plugin):
    depends_on = ['s2']
    provides = ['acc_s2_threshold']
    parameters = ('s2_threshold',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, s2):
        return key, jnp.where(s2 > parameters['s2_threshold'], 1., 0)


@export
@appletree.takes_config(
    SigmaMap(name='s1_eff_3f',
        default=[
            '3fold_recon_eff.json',
            '3fold_recon_eff.json',
            '3fold_recon_eff.json'],
        help='3fold S1 reconstruction efficiency'),
)
class S1ReconEff(Plugin):
    depends_on = ['num_s1_phd']
    provides = ['acc_s1_recon_eff']
    parameters = ('s1_eff_3f_sigma',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_s1_phd):
        acc_s1_recon_eff = self.s1_eff_3f.apply(
            num_s1_phd, parameters)
        acc_s1_recon_eff = jnp.clip(acc_s1_recon_eff, 0., 1.)
        return key, acc_s1_recon_eff


@export
@appletree.takes_config(
    SigmaMap(name='s1_cut_acc',
        default=[
            's1_cut_acc.json',
            's1_cut_acc.json',
            's1_cut_acc.json'],
        help='S1 cut acceptance'),
)
class S1CutAccept(Plugin):
    depends_on = ['s1']
    provides = ['cut_acc_s1']
    parameters = ('s1_cut_acc_sigma',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, s1):
        cut_acc_s1 = self.s1_cut_acc.apply(s1, parameters)
        cut_acc_s1 = jnp.clip(cut_acc_s1, 0., 1.)
        return key, cut_acc_s1


@export
@appletree.takes_config(
    SigmaMap(name='s2_cut_acc',
        default=[
            's2_cut_acc.json',
            's2_cut_acc.json',
            's2_cut_acc.json'],
        help='S2 cut acceptance'),
)
class S2CutAccept(Plugin):
    depends_on = ['s2']
    provides = ['cut_acc_s2']
    parameters = ('s2_cut_acc_sigma',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, s2):
        cut_acc_s2 = self.s2_cut_acc.apply(s2, parameters)
        cut_acc_s2 = jnp.clip(cut_acc_s2, 0., 1.)
        return key, cut_acc_s2


@export
class Eff(Plugin):
    depends_on = ['acc_s2_threshold', 'acc_s1_recon_eff', 'cut_acc_s1', 'cut_acc_s2']
    provides = ['eff']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, acc_s2_threshold, acc_s1_recon_eff, cut_acc_s1, cut_acc_s2):
        eff = acc_s2_threshold * acc_s1_recon_eff * cut_acc_s1 * cut_acc_s2
        return key, eff



from jax import numpy as jnp
from jax import jit
from functools import partial

import appletree
from appletree import randgen
from appletree.plugin import Plugin
from appletree.config import Map
from appletree.utils import exporter

export, __all__ = exporter(export_self=False)


@export
@appletree.takes_config(
    Map(name='s1_lce',
        default='s1_correction_map.json',
        help='S1 light collection efficiency correction'),
)
class S1Correction(Plugin):
    depends_on = ['rec_x', 'rec_y', 'rec_z']
    provides = ['s1_correction']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, rec_x, rec_y, rec_z):
        pos = jnp.stack([rec_x, rec_y, rec_z]).T
        s1_correction = self.s1_lce.apply(pos)
        return key, s1_correction


@export
@appletree.takes_config(
    Map(name='s2_lce',
        default='s2_correction_map.json',
        help='S2 light collection efficiency correction'),
)
class S2Correction(Plugin):
    depends_on = ['rec_x', 'rec_y']
    provides = ['s2_correction']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, rec_x, rec_y):
        pos = jnp.stack([rec_x, rec_y]).T
        s2_correction = self.s2_lce.apply(pos)
        return key, s2_correction


@export
class PhotonDetection(Plugin):
    depends_on = ['num_photon', 's1_correction']
    provides = ['num_s1_phd']
    parameters = ('g1', 'p_dpe')

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_photon, s1_correction):
        g1_true_no_dpe = jnp.clip(parameters['g1'] * s1_correction / (1. + parameters['p_dpe']), 0, 1.)
        key, num_s1_phd = randgen.binomial(key, g1_true_no_dpe, num_photon)
        return key, num_s1_phd


@export
class S1PE(Plugin):
    depends_on = ['num_s1_phd']
    provides = ['num_s1_pe']
    parameters = ('p_dpe',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_s1_phd):
        key, num_s1_dpe = randgen.binomial(key, parameters['p_dpe'], num_s1_phd)
        num_s1_pe = num_s1_dpe + num_s1_phd
        return key, num_s1_pe


@export
@appletree.takes_config(
    Map(name='elife',
        default='elife.json',
        help='Electron lifetime correction'),
)
class DriftLoss(Plugin):
    depends_on = ['z']
    provides = ['drift_survive_prob']
    parameters = ('drift_velocity',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, z):
        key, p = randgen.uniform(key, 0, 1., shape=jnp.shape(z))
        lifetime = self.elife.apply(p)
        drift_survive_prob = jnp.exp(- jnp.abs(z) / parameters['drift_velocity'] / lifetime)
        return key, drift_survive_prob


@export
class ElectronDrifted(Plugin):
    depends_on = ['num_electron', 'drift_survive_prob']
    provides = ['num_electron_drifted']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_electron, drift_survive_prob):
        key, num_electron_drifted = randgen.binomial(key, drift_survive_prob, num_electron)
        return key, num_electron_drifted


@export
class S2PE(Plugin):
    depends_on = ['num_electron_drifted', 's2_correction']
    provides = ['num_s2_pe']
    parameters = ('g2', 'gas_gain')

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_electron_drifted, s2_correction):
        extraction_eff = parameters['g2'] / parameters['gas_gain']
        g2_true = parameters['g2'] * s2_correction
        gas_gain_true = g2_true / extraction_eff

        key, num_electron_extracted = randgen.binomial(key, extraction_eff, num_electron_drifted)

        mean_s2_pe = num_electron_extracted * gas_gain_true
        key, num_s2_pe = randgen.truncate_normal(key, mean_s2_pe, jnp.sqrt(mean_s2_pe), vmin=0)

        return key, num_s2_pe



from jax import numpy as jnp
from jax import jit
from functools import partial

import appletree
from appletree import randgen
from appletree.plugin import Plugin
from appletree.config import Map
from appletree.utils import exporter

export, __all__ = exporter(export_self=False)


@export
@appletree.takes_config(
    Map(name='ly_median',
        default='nr_ly_median.json',
        help='Light yield curve from NESTv2'),
)
class LightYield(Plugin):
    depends_on = ['energy']
    provides = ['light_yield']
    parameters = ('t_ly',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy):
        light_yield = self.ly_median.apply(energy) * (1. + parameters['t_ly'])
        light_yield = jnp.clip(light_yield, 0, jnp.inf)
        return key, light_yield


@export
class NumberPhoton(Plugin):
    depends_on = ['energy', 'light_yield']
    provides = ['num_photon']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy, light_yield):
        key, num_photon = randgen.poisson(key, light_yield * energy)
        return key, num_photon


@export
@appletree.takes_config(
    Map(name='qy_median',
        default='nr_qy_median.json',
        help='Charge yield curve from NESTv2'),
)
class ChargeYield(Plugin):
    depends_on = ['energy']
    provides = ['charge_yield']
    parameters = ('t_qy',)

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy):
        charge_yield = self.qy_median.apply(energy) * (1. + parameters['t_qy'])
        charge_yield = jnp.clip(charge_yield, 0, jnp.inf)
        return key, charge_yield


@export
class NumberElectron(Plugin):
    depends_on = ['energy', 'charge_yield']
    provides = ['num_electron']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, energy, charge_yield):
        key, num_electron = randgen.poisson(key, charge_yield * energy)
        return key, num_electron



__version__ = '0.1.0'

from . import utils
from .utils import *

from . import hist
from .hist import *

from . import interpolation
from .interpolation import *

from . import config
from .config import *

from . import parameter
from .parameter import *

from . import randgen
from .randgen import *

from . import share
from .share import *

from . import plugins

from . import plugin
from .plugin import *

from . import components
from .components import *

from . import component
from .component import *

from . import likelihood
from .likelihood import *

from . import contexts
from .contexts import *

from . import context
from .context import *



from functools import partial

from jax import jit
from jax import numpy as jnp

import appletree
from appletree import randgen
from appletree.plugin import Plugin
from appletree.config import Constant, Map
from appletree.utils import exporter

export, __all__ = exporter()


@export
@appletree.takes_config(
    Constant(name='lower_energy',
        type=float,
        default=0.01,
        help='Energy lower limit simulated in uniformly distribution'),
    Constant(name='upper_energy',
        type=float,
        default=20.,
        help='Energy upper limit simulated in uniformly distribution'),
)
class UniformEnergySpectra(Plugin):
    depends_on = ['batch_size']
    provides = ['energy']

    @partial(jit, static_argnums=(0, 3))
    def simulate(self, key, parameters, batch_size):
        key, energy = randgen.uniform(
            key,
            self.lower_energy.value,
            self.upper_energy.value,
            shape=(batch_size, ),
        )
        return key, energy


@export
@appletree.takes_config(
    Map(name='energy_spectrum',
        default='nr_spectrum.json',
        help='Recoil energy spectrum'),
)
class FixedEnergySpectra(Plugin):
    depends_on = ['batch_size']
    provides = ['energy']

    @partial(jit, static_argnums=(0, 3))
    def simulate(self, key, parameters, batch_size):
        key, p = randgen.uniform(key, 0, 1., shape=(batch_size, ))
        energy = self.energy_spectrum.apply(p)
        return key, energy


@export
@appletree.takes_config(
    Constant(name='mono_energy',
        type=float,
        default=2.82,
        help='Mono energy delta function'),
)
class MonoEnergySpectra(Plugin):
    depends_on = ['batch_size']
    provides = ['energy']

    # default energy is Ar37 K shell

    @partial(jit, static_argnums=(0, 3))
    def simulate(self, key, parameters, batch_size):
        energy = jnp.full(batch_size, self.mono_energy.value)
        return key, energy


@export
@appletree.takes_config(
    Constant(name='z_min',
        type=float,
        default=-133.97,
        help='Z lower limit simulated in uniformly distribution'),
    Constant(name='z_max',
        type=float,
        default=-13.35,
        help='Z upper limit simulated in uniformly distribution'),
    Constant(name='r_max',
        type=float,
        default=60.,
        help='Radius upper limit simulated in uniformly distribution'),
)
class PositionSpectra(Plugin):
    depends_on = ['batch_size']
    provides = ['x', 'y', 'z']

    @partial(jit, static_argnums=(0, 3))
    def simulate(self, key, parameters, batch_size):
        key, z = randgen.uniform(key, self.z_min.value, self.z_max.value, shape=(batch_size, ))
        key, r2 = randgen.uniform(key, 0, self.r_max.value**2, shape=(batch_size, ))
        key, theta = randgen.uniform(key, 0, 2*jnp.pi, shape=(batch_size, ))

        r = jnp.sqrt(r2)
        x = r * jnp.cos(theta)
        y = r * jnp.sin(theta)
        return key, x, y, z



from functools import partial

from jax import numpy as jnp
from jax import jit, lax

from appletree.utils import exporter

export, __all__ = exporter(export_self=False)

FLOAT_POS_MIN = jnp.finfo(jnp.float32).tiny
FLOAT_POS_MAX = jnp.finfo(jnp.float32).max


@jit
def _L2_dist2(pos1, pos2):
    """Calculate L2 distance between pos1 and pos2.

    :param pos1: array with shape (N, D)
    :param pos2: array with shape (M, D)
    :return: L2 distance squared with shape (N, M)
    """
    dr = jnp.expand_dims(pos1, axis=1) - jnp.expand_dims(pos2, axis=0)
    return jnp.sum(dr * dr, axis=-1)


@export
@partial(jit, static_argnums=(3,))
def map_interpolator_knn(pos, ref_pos, ref_val, k=3):
    """Inverse distance weighting average as interpolation using KNN.

    :param pos: array with shape (N, D), as the points to be interpolated.
    :param ref_pos: array with shape (M, D), as the reference points.
    :param ref_val: array with shape (M, ), as the reference values.
    :return: interpolated values with shape (N, ), weighted by the inverse of
    the distance to k nearest neighbors.
    """
    pos = jnp.asarray(pos)
    ref_pos = jnp.asarray(ref_pos)
    ref_val = jnp.asarray(ref_val)

    dr2 = -_L2_dist2(pos, ref_pos)
    dr2, ind = lax.top_k(dr2, k)
    weights = 1.0 / jnp.clip(jnp.sqrt(-dr2), 1e-6, float('inf'))
    val = jnp.take(ref_val, ind)
    val = jnp.sum(val * weights, axis=1) / jnp.sum(weights, axis=1)

    return val


@export
@jit
def curve_interpolator(pos, ref_pos, ref_val):
    """Inverse distance weighting average as interpolation using KNN (K=2) for 1D map.

    :param pos: array with shape (N, ), as the points to be interpolated.
    :param ref_pos: array with shape (M, ), as the reference points.
    :param ref_val: array with shape (M, ), as the reference values.
    :return: interpolated values with shape (N, ), weighted by the inverse of
        the distance to k nearest neighbors.
    """
    right = jnp.searchsorted(ref_pos, pos)
    left = right - 1

    right = jnp.clip(right, 0, len(ref_pos)-1)
    left = jnp.clip(left, 0, len(ref_pos)-1)

    val_right = ref_val[right]
    val_left = ref_val[left]

    dist_right = jnp.abs(pos - ref_pos[right])
    dist_left = jnp.abs(pos - ref_pos[left])

    val = jnp.where(
        (dist_right+dist_left) > 0,
        (val_right*dist_left+val_left*dist_right)/(dist_right+dist_left),
        val_right,
    )
    return val


@export
@jit
def map_interpolator_regular_binning_1d(pos, ref_pos_lowers, ref_pos_uppers, ref_val):
    """Inverse distance weighting average as 1D interpolation using KNN(K=2).
    A uniform mesh grid binning is assumed.

    :param pos: array with shape (N, ), positions at which the interp is calculated.
    :param ref_pos_lowers: array with shape (1, ), the lower edges of the binning on each dimension.
    :param ref_pos_uppers: array with shape (1, ), the upper edges of the binning on each dimension.
    :param ref_val: array with shape (M1, ), map values.
    """
    ref_pos = jnp.linspace(ref_pos_lowers, ref_pos_uppers, len(ref_val))
    val = curve_interpolator(pos, ref_pos, ref_val)

    return val


@export
@jit
def map_interpolator_regular_binning_2d(pos, ref_pos_lowers, ref_pos_uppers, ref_val):
    """Inverse distance weighting average as 2D interpolation using KNN(K=4).
    A uniform mesh grid binning is assumed.

    :param pos: array with shape (N, 2), positions at which the interp is calculated.
    :param ref_pos_lowers: array with shape (2, ), the lower edges of the binning on each dimension.
    :param ref_pos_uppers: array with shape (2, ), the upper edges of the binning on each dimension.
    :param ref_val: array with shape (M1, M2), map values.
    """
    num_bins = jnp.asarray(jnp.shape(ref_val))
    bin_sizes = (ref_pos_uppers - ref_pos_lowers) / (num_bins - 1)
    num_bins = num_bins[jnp.newaxis, :]
    bin_sizes = bin_sizes[jnp.newaxis, :]

    ind1 = jnp.floor((pos - ref_pos_lowers) / bin_sizes)
    ind1 = jnp.clip(ind1, a_min=0, a_max=num_bins-1)
    ind1 = jnp.asarray(ind1, dtype=int)
    ind2 = ind1.at[:, 0].add(1)
    ind3 = ind1.at[:, 1].add(1)
    ind4 = ind2.at[:, 1].add(1)

    val1 = ref_val[ind1[:, 0], ind1[:, 1]]
    val2 = ref_val[ind2[:, 0], ind2[:, 1]]
    val3 = ref_val[ind3[:, 0], ind3[:, 1]]
    val4 = ref_val[ind4[:, 0], ind4[:, 1]]

    ref_pos1 = ref_pos_lowers + bin_sizes * ind1
    ref_pos2 = ref_pos_lowers + bin_sizes * ind2
    ref_pos3 = ref_pos_lowers + bin_sizes * ind3
    ref_pos4 = ref_pos_lowers + bin_sizes * ind4

    dr1 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos1 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr2 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos2 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr3 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos3 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr4 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos4 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)

    val = (val1/dr1+val2/dr2+val3/dr3+val4/dr4)
    val /= (1./dr1+1./dr2+1./dr3+1./dr4)

    return val


@export
@jit
def map_interpolator_regular_binning_3d(pos, ref_pos_lowers, ref_pos_uppers, ref_val):
    """Inverse distance weighting average as 3D interpolation using KNN(K=8).
    A uniform mesh grid binning is assumed.

    :param pos: array with shape (N, 3), positions at which the interp is calculated.
    :param ref_pos_lowers: array with shape (3, ), the lower edges of the binning on each dimension.
    :param ref_pos_uppers: array with shape (3, ), the upper edges of the binning on each dimension.
    :param ref_val: array with shape (M1, M2, M3), map values.
    """
    num_bins = jnp.asarray(jnp.shape(ref_val))
    bin_sizes = (ref_pos_uppers - ref_pos_lowers) / (num_bins - 1)
    num_bins = num_bins[jnp.newaxis, :]
    bin_sizes = bin_sizes[jnp.newaxis, :]

    ind1 = jnp.floor((pos - ref_pos_lowers) / bin_sizes)
    ind1 = jnp.clip(ind1, a_min=0, a_max=num_bins-1)
    ind1 = jnp.asarray(ind1, dtype=int)
    ind2 = ind1.at[:, 0].add(1)
    ind3 = ind1.at[:, 1].add(1)
    ind4 = ind1.at[:, 2].add(1)
    ind5 = ind2.at[:, 1].add(1)
    ind6 = ind3.at[:, 2].add(1)
    ind7 = ind4.at[:, 0].add(1)
    ind8 = ind7.at[:, 1].add(1)

    val1 = ref_val[ind1[:, 0], ind1[:, 1], ind1[:, 2]]
    val2 = ref_val[ind2[:, 0], ind2[:, 1], ind2[:, 2]]
    val3 = ref_val[ind3[:, 0], ind3[:, 1], ind3[:, 2]]
    val4 = ref_val[ind4[:, 0], ind4[:, 1], ind4[:, 2]]
    val5 = ref_val[ind5[:, 0], ind5[:, 1], ind5[:, 2]]
    val6 = ref_val[ind6[:, 0], ind6[:, 1], ind6[:, 2]]
    val7 = ref_val[ind7[:, 0], ind7[:, 1], ind7[:, 2]]
    val8 = ref_val[ind8[:, 0], ind8[:, 1], ind8[:, 2]]

    ref_pos1 = ref_pos_lowers + bin_sizes * ind1
    ref_pos2 = ref_pos_lowers + bin_sizes * ind2
    ref_pos3 = ref_pos_lowers + bin_sizes * ind3
    ref_pos4 = ref_pos_lowers + bin_sizes * ind4
    ref_pos5 = ref_pos_lowers + bin_sizes * ind5
    ref_pos6 = ref_pos_lowers + bin_sizes * ind6
    ref_pos7 = ref_pos_lowers + bin_sizes * ind7
    ref_pos8 = ref_pos_lowers + bin_sizes * ind8

    dr1 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos1 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr2 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos2 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr3 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos3 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr4 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos4 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr5 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos5 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr6 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos6 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr7 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos7 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)
    dr8 = jnp.clip(jnp.sqrt(jnp.sum((ref_pos8 - pos)**2, axis=-1)), a_min=FLOAT_POS_MIN, a_max=FLOAT_POS_MAX)

    val = (val1/dr1+val2/dr2+val3/dr3+val4/dr4+val5/dr5+val6/dr6+val7/dr7+val8/dr8)
    val /= (1./dr1+1./dr2+1./dr3+1./dr4+1./dr5+1./dr6+1./dr7+1./dr8)

    return val



import appletree as apt

from appletree.component import ComponentSim


class ERBand(ComponentSim):
    norm_type = 'on_pdf'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.register(apt.plugins.common.UniformEnergySpectra)
        self.register(apt.plugins.common.PositionSpectra)
        self.register_all(apt.plugins.er_microphys)
        self.register_all(apt.plugins.detector)
        self.register_all(apt.plugins.reconstruction)
        self.register_all(apt.plugins.efficiency)


class ERPeak(ComponentSim):
    norm_type = 'on_pdf'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.register(apt.plugins.common.MonoEnergySpectra)
        self.register(apt.plugins.common.PositionSpectra)
        self.register_all(apt.plugins.er_microphys)
        self.register_all(apt.plugins.detector)
        self.register_all(apt.plugins.reconstruction)
        self.register_all(apt.plugins.efficiency)



from jax import jit
from functools import partial

from jax import numpy as jnp

import appletree
from appletree import randgen
from appletree.config import Map
from appletree.plugin import Plugin
from appletree.utils import exporter

export, __all__ = exporter(export_self=False)


@export
@appletree.takes_config(
    Map(name='posrec_reso',
        default='posrec_reso.json',
        help='Position reconstruction resolution'),
)
class PositionRecon(Plugin):
    depends_on = ['x', 'y', 'z', 'num_electron_drifted']
    provides = ['rec_x', 'rec_y', 'rec_z', 'rec_r']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, x, y, z, num_electron_drifted):
        std = self.posrec_reso.apply(num_electron_drifted)
        std /= jnp.sqrt(2)
        mean = jnp.zeros_like(num_electron_drifted)
        key, delta_x = randgen.normal(key, mean, std)
        key, delta_y = randgen.normal(key, mean, std)
        rec_x = x + delta_x
        rec_y = y + delta_y
        rec_z = z
        rec_r = jnp.sqrt(rec_x**2 + rec_y**2)
        return key, rec_x, rec_y, rec_z, rec_r


@export
@appletree.takes_config(
    Map(name='s1_bias_3f',
        default='s1_bias_3f.json',
        help='3fold S1 reconstruction bias'),
    Map(name='s1_smear_3f',
        default='s1_smearing_3f.json',
        help='3fold S1 reconstruction smearing'),
)
class S1(Plugin):
    depends_on = ['num_s1_phd', 'num_s1_pe']
    provides = ['s1']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_s1_phd, num_s1_pe):
        mean = self.s1_bias_3f.apply(num_s1_phd)
        std = self.s1_smear_3f.apply(num_s1_phd)
        key, bias = randgen.normal(key, mean, std)
        s1 = num_s1_pe * (1. + bias)
        return key, s1


@export
@appletree.takes_config(
    Map(name='s2_bias',
        default='s2_bias.json',
        help='S2 reconstruction bias'),
    Map(name='s2_smear',
        default='s2_smearing.json',
        help='S2 reconstruction smearing'),
)
class S2(Plugin):
    depends_on = ['num_s2_pe']
    provides = ['s2']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, num_s2_pe):
        mean = self.s2_bias.apply(num_s2_pe)
        std = self.s2_smear.apply(num_s2_pe)
        key, bias = randgen.normal(key, mean, std)
        s2 = num_s2_pe * (1. + bias)
        return key, s2


@export
class cS1(Plugin):
    depends_on = ['s1', 's1_correction']
    provides = ['cs1']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, s1, s1_correction):
        cs1 = s1 / s1_correction
        return key, cs1


@export
class cS2(Plugin):
    depends_on = ['s2', 's2_correction', 'drift_survive_prob']
    provides = ['cs2']

    @partial(jit, static_argnums=(0, ))
    def simulate(self, key, parameters, s2, s2_correction, drift_survive_prob):
        cs2 = s2 / s2_correction / drift_survive_prob
        return key, cs2



import setuptools


def open_requirements(path):
    with open(path) as f:
        requires = [
            r.split('/')[-1] if r.startswith('git+') else r
            for r in f.read().splitlines()]
    return requires


requires = open_requirements('requirements.txt')

with open('README.md') as file:
    readme = file.read()

with open('HISTORY.md') as file:
    history = file.read()

setuptools.setup(
    name='appletree',
    version='0.1.0',
    description='A high-Performance Program simuLatEs and fiTs REsponse of xEnon.',
    author='Appletree contributors, the XENON collaboration',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    setup_requires=['pytest-runner'],
    install_requires=requires,
    python_requires='>=3.8',
    extras_require={
        'doc': [],
        'test': [
            'pytest',
            'flake8',
        ],
    },
    packages=setuptools.find_packages(),
    url="https://github.com/XENONnT/appletree",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    zip_safe=False,
)



_cached_configs = dict()
_cached_functions = dict()


def set_global_config(configs):
    """Set new global configuration options

    :param configs: dict, configuration file name or dictionary
    """
    from appletree.utils import get_file_path

    for k, v in configs.items():
        if isinstance(v, (float, int, list)):
            _cached_configs.update({k: v})
        elif isinstance(v, str):
            file_path = get_file_path(v)
            _cached_configs.update({k: file_path})
        elif isinstance(v, dict):
            file_path_dict = {}
            for kk, vv in v.items():
                if isinstance(vv, (float, int, list)):
                    file_path_dict[kk] = vv
                elif isinstance(vv, str):
                    file_path_dict[kk] = get_file_path(vv)
            _cached_configs.update({k: file_path_dict})
        else:
            raise NotImplementedError



import os
import re
import json
from warnings import warn
import pkg_resources
from time import time

import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib.patches import Rectangle
from matplotlib import pyplot as plt

import GOFevaluation
from appletree.share import _cached_configs

NT_AUX_INSTALLED = False
try:
    import ntauxfiles
    NT_AUX_INSTALLED = True
except ImportError:
    pass


def exporter(export_self=False):
    """Export utility modified from https://stackoverflow.com/a/41895194
    Returns export decorator, __all__ list
    """
    all_ = []
    if export_self:
        all_.append('exporter')

    def decorator(obj):
        all_.append(obj.__name__)
        return obj

    return decorator, all_


export, __all__ = exporter(export_self=True)


@export
def use_xenon_plot_style():
    """Set matplotlib plot style."""
    params = {
        'font.family': 'serif',
        'font.size': 24, 'axes.titlesize': 24,
        'axes.labelsize': 24, 'axes.linewidth': 2,
        # ticks
        'xtick.labelsize': 22, 'ytick.labelsize': 22, 'xtick.major.size': 16, 'xtick.minor.size': 8,
        'ytick.major.size': 16, 'ytick.minor.size': 8, 'xtick.major.width': 2, 'xtick.minor.width': 2,
        'ytick.major.width': 2, 'ytick.minor.width': 2, 'xtick.direction': 'in', 'ytick.direction': 'in',
        # markers
        'lines.markersize': 12, 'lines.markeredgewidth': 3, 'errorbar.capsize': 8, 'lines.linewidth': 3,
        'savefig.bbox': 'tight', 'legend.fontsize': 24,
        'backend': 'Agg', 'mathtext.fontset': 'dejavuserif', 'legend.frameon': False,
        # figure
        'figure.facecolor': 'w',
        'figure.figsize': (12, 8),
        # pad
        'axes.labelpad': 12,
        # ticks
        'xtick.major.pad': 6, 'xtick.minor.pad': 6,
        'ytick.major.pad': 3.5, 'ytick.minor.pad': 3.5,
        # colormap
    }
    plt.rcParams.update(params)


@export
def load_data(file_name: str):
    """Load data from file. The suffix can be ".csv", ".pkl"."""
    file_name = get_file_path(file_name)
    fmt = file_name.split('.')[-1]
    if fmt == 'csv':
        data = pd.read_csv(file_name)
    elif fmt == 'pkl':
        data = pd.read_pickle(file_name)
    else:
        raise ValueError(f'unsupported file format {fmt}!')
    return data


@export
def load_json(file_name: str):
    """Load data from json file."""
    with open(get_file_path(file_name), 'r') as file:
        data = json.load(file)
    return data


@export
def camel_to_snake(x):
    """Convert x from CamelCase to snake_case,
    from https://stackoverflow.com/questions/1175208
    """
    x = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', x)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', x).lower()


@export
def _get_abspath(file_name):
    """Get the abspath of the file. Raise FileNotFoundError when not found in any subfolder"""
    for sub_dir in ('maps', 'data', 'parameters', 'instructs'):
        p = os.path.join(_package_path(sub_dir), file_name)
        if os.path.exists(p):
            return p
    raise FileNotFoundError(f'Cannot find {file_name}')


def _package_path(sub_directory):
    """Get the abs path of the requested sub folder"""
    return pkg_resources.resource_filename('appletree', f'{sub_directory}')


@export
def get_file_path(fname):
    """Find the full path to the resource file
    Try 5 methods in the following order

    #. fname begin with '/', return absolute path
    #. url_base begin with '/', return url_base + name
    #. can get file from _get_abspath, return appletree internal file path
    #. can be found in local installed ntauxfiles, return ntauxfiles absolute path
    #. can be downloaded from MongoDB, download and return cached path
    """
    # 1. From absolute path
    # Usually Config.default is a absolute path
    if fname.startswith('/'):
        return fname

    # 2. From local folder
    # Use url_base as prefix
    if 'url_base' in _cached_configs.keys():
        url_base = _cached_configs['url_base']

        if url_base.startswith('/'):
            fpath = os.path.join(url_base, fname)
            if os.path.exists(fpath):
                warn(f'Load {fname} successfully from {fpath}')
                return fpath

    # 3. From appletree internal files
    try:
        return _get_abspath(fname)
    except FileNotFoundError:
        pass

    # 4. From local installed ntauxfiles
    if NT_AUX_INSTALLED:
        # You might want to use this, for example if you are a developer
        if fname in ntauxfiles.list_private_files():
            fpath = ntauxfiles.get_abspath(fname)
            warn(f'Load {fname} successfully from {fpath}')
            return fpath

    # 5. From MongoDB
    try:
        import straxen
        # https://straxen.readthedocs.io/en/latest/config_storage.html
        # downloading-xenonnt-files-from-the-database  # noqa

        # we need to add the straxen.MongoDownloader() in this
        # try: except NameError: logic because the NameError
        # gets raised if we don't have access to utilix.
        downloader = straxen.MongoDownloader()
        # FileNotFoundError, ValueErrors can be raised if we
        # cannot load the requested config
        fpath = downloader.download_single(fname)
        warn(f'Loading {fname} from mongo downloader to {fpath}')
        return fname  # Keep the name and let get_resource do its thing
    except (FileNotFoundError, ValueError, NameError, AttributeError):
        warn(f'Mongo downloader not possible or does not have {fname}')

    # raise error when can not find corresponding file
    raise RuntimeError(f'Can not find {fname}, please check your file system')


@export
def timeit(indent=""):
    """Use timeit as a decorator.
    It will print out the running time of the decorated function.
    """
    def _timeit(func, indent):
        name = func.__name__

        def _func(*args, **kwargs):
            print(indent + f' Function <{name}> starts.')
            start = time()
            res = func(*args, **kwargs)
            time_ = (time() - start) * 1e3
            print(indent + f' Function <{name}> ends! Time cost = {time_:.2f} msec.')
            return res

        return _func
    if isinstance(indent, str):
        return lambda func: _timeit(func, indent)
    else:
        return _timeit(indent, "")


@export
def set_gpu_memory_usage(fraction=0.3):
    """Set GPU memory usage.
    See more on https://jax.readthedocs.io/en/latest/gpu_memory_allocation.html
    """
    if fraction > 1:
        fraction = 1
    if fraction <= 0:
        raise ValueError("fraction must be positive!")
    os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = f"{fraction:.2f}"


@export
def get_equiprob_bins_2d(data,
                         n_partitions,
                         order = (0, 1),
                         x_clip = (-np.inf, +np.inf),
                         y_clip = (-np.inf, +np.inf),
                         which_np = np):
    """Get 2D equiprobable binning edges.

    :param data: array with shape (N, 2).
    :param n_partitions: [M1, M2] where M1 M2 are the number of bins on each dimension.
    :param x_clip: lower and upper binning edges on the 0th dimension.
        Data outside the x_clip will be dropped.
    :param y_clip: lower and upper binning edges on the 1st dimension.
        Data outside the y_clip will be dropped.
    :param which_np: can be numpy or jax.numpy, determining the returned array type.
    """
    mask = (data[:, 0] > x_clip[0])
    mask &= (data[:, 0] < x_clip[1])
    mask &= (data[:, 1] > y_clip[0])
    mask &= (data[:, 1] < y_clip[1])

    x_bins, y_bins = GOFevaluation.utils.get_equiprobable_binning(
        data[mask],
        n_partitions,
        order = order,
    )
    # To be strict, clip the inf(s)
    x_bins = np.clip(x_bins, *x_clip)
    y_bins = np.clip(y_bins, *y_clip)

    return which_np.array(x_bins), which_np.array(y_bins)


@export
def plot_irreg_histogram_2d(bins_x, bins_y, hist, **kwargs):
    """Plot histogram defined by irregular binning.

    :param bins_x: array with shape (M1, )
    :param bins_y: array with shape (M1-1, M2)
    :param hist: array with shape (M1-1, M2-1)
    :param density: boolean.
    """
    hist = np.asarray(hist)
    bins_x = np.asarray(bins_x)
    bins_y = np.asarray(bins_y)

    density = kwargs.get('density', False)
    cmap = mpl.cm.get_cmap("RdBu_r")

    loc = []
    width = []
    height = []
    area = []
    n = []

    for i, _ in enumerate(hist):
        for j, _ in enumerate(hist[i]):
            x_lower = bins_x[i]
            x_upper = bins_x[i+1]
            y_lower = bins_y[i, j]
            y_upper = bins_y[i, j+1]

            loc.append((x_lower, y_lower))
            width.append(x_upper - x_lower)
            height.append(y_upper - y_lower)
            area.append((x_upper - x_lower) * (y_upper - y_lower))
            n.append(hist[i, j])

    loc = np.asarray(loc)
    width = np.asarray(width)
    height = np.asarray(height)
    area = np.asarray(area)
    n = np.asarray(n)

    if density:
        norm = mpl.colors.Normalize(
            vmin=kwargs.get('vmin', np.min(n/area)),
            vmax=kwargs.get('vmax', np.max(n/area)),
            clip=False,
        )
    else:
        norm = mpl.colors.Normalize(
            vmin=kwargs.get('vmin', np.min(n)),
            vmax=kwargs.get('vmax', np.max(n)),
            clip=False,
        )

    ax = plt.subplot()
    for i, _ in enumerate(loc):
        c = n[i]/area[i] if density else n[i]
        rec = Rectangle(
            loc[i],
            width[i],
            height[i],
            facecolor=cmap(norm(c)),
            edgecolor='k',
        )
        ax.add_patch(rec)

    fig = plt.gcf()
    fig.colorbar(
        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
        ax=ax,
        label=('# events / bin size' if density else '# events'),
    )

    ax.set_xlim(np.min(bins_x), np.max(bins_x))
    ax.set_ylim(np.min(bins_y), np.max(bins_y))

    return ax


@export
def add_spaces(x):
    """Add four spaces to every line in x
    This is needed to make html raw blocks in rst format correctly
    """
    y = ''
    if isinstance(x, str):
        x = x.split('\n')
    for q in x:
        y += '    ' + q
    return y


@export
def tree_to_svg(graph_tree, save_as='data_types', view=True):
    """
    Where to save this node

    :param graph_tree: Digraph instance
    :param save_as: str, file name
    :param view: bool, Open the rendered result with the default application.
    """
    graph_tree.render(save_as, view=view)
    with open(f'{save_as}.svg', mode='r') as f:
        svg = add_spaces(f.readlines()[5:])
    os.remove(save_as)
    return svg


@export
def add_deps_to_graph_tree(component,
                           graph_tree,
                           data_names: list = ['cs1', 'cs2', 'eff'],
                           _seen = None):
    """
    Recursively add nodes to graph base on plugin.deps

    :param context: Context instance
    :param graph_tree: Digraph instance
    :param data_names: Data type name
    :param _seen: list or None, the seen data_name should not be plot
    """
    if _seen is None:
        _seen = []
    for data_name in data_names:
        if data_name in _seen:
            continue

        # Add new one
        graph_tree.node(data_name,
                        style='filled',
                        href='#' + data_name.replace('_', '-'),
                        fillcolor='white')
        if data_name == 'batch_size':
            continue
        dep_plugin = component._plugin_class_registry[data_name]
        for dep in dep_plugin.depends_on:
            graph_tree.edge(data_name, dep)
            graph_tree, _seen = add_deps_to_graph_tree(
                component,
                graph_tree,
                dep_plugin.depends_on,
                _seen,
            )
        _seen.append(data_name)
    return graph_tree, _seen


@export
def add_plugins_to_graph_tree(component,
                              graph_tree,
                              data_names: list = ['cs1', 'cs2', 'eff'],
                              _seen = None,
                              with_data_names=False):
    """
    Recursively add nodes to graph base on plugin.deps

    :param context: Context instance
    :param graph_tree: Digraph instance
    :param data_names: Data type name
    :param _seen: list or None, the seen data_name should not be plot
    :param with_data_names: bool, whether plot even with messy data_names
    """
    if _seen is None:
        _seen = []
    for data_name in data_names:
        if data_name == 'batch_size':
            continue

        plugin = component._plugin_class_registry[data_name]
        plugin_name = plugin.__name__
        if plugin_name in _seen:
            continue

        # Add new one
        label = f'{plugin_name}'
        if with_data_names:
            label += f"\n{', '.join(plugin.depends_on)}\n{', '.join(plugin.provides)}"
        graph_tree.node(plugin_name,
                        label=label,
                        style='filled',
                        href='#' + plugin_name.replace('_', '-'),
                        fillcolor='white')

        for dep in plugin.depends_on:
            if dep == 'batch_size':
                continue
            dep_plugin = component._plugin_class_registry[dep]
            graph_tree.edge(plugin_name, dep_plugin.__name__)
            graph_tree, _seen = add_plugins_to_graph_tree(
                component,
                graph_tree,
                plugin.depends_on,
                _seen,
            )
        _seen.append(plugin_name)
    return graph_tree, _seen


@export
def add_extensions(module1, module2, base, force=False):
    """
    Add subclasses of module2 to module1

    When ComponentSim compiles the dependency tree,
    it will search in the appletree.plugins module for Plugin(as attributes).
    When building Likelihood, it will also search for corresponding Component(s)
    specified in the instructions(e.g. NRBand).

    So we need to assign the attributes before compilation.
    These plugins are mostly user defined.
    """
    # Assign the module2 as attribute of module1
    is_exists = module2.__name__ in dir(module1)
    if is_exists and not force:
        raise ValueError(
            f'{module2.__name__} already existed in {module1.__name__}, '
            f'do not re-register a module with same name',
        )
    else:
        if is_exists:
            print(f'You have forcibly registered {module2.__name__} to {module1.__name__}')
        setattr(module1, module2.__name__.split('.')[-1], module2)

    # Iterate the module2 and assign the single Plugin(s) as attribute(s)
    for x in dir(module2):
        x = getattr(module2, x)
        if not isinstance(x, type(type)):
            continue
        _add_extension(module1, x, base, force=force)


def _add_extension(module, subclass, base, force=False):
    """
    Add subclass to module
    Skip the class when it is base class.

    It is no allowed to assign a class which has same name to an already assigned class.
    We do not allowed class name covering!
    Please change the name of your class when Error shows itself.
    """
    if getattr(subclass, '_' + subclass.__name__ + '__is_base', False):
        return

    if issubclass(subclass, base) and subclass != base:
        is_exists = subclass.__name__ in dir(module)
        if is_exists and not force:
            raise ValueError(
                f'{subclass.__name__} already existed in {module.__name__}, '
                f'do not re-register a {base.__name__} with same name',
            )
        else:
            if is_exists:
                print(f'You have forcibly registered {subclass.__name__} to {module.__name__}')
            setattr(module, subclass.__name__, subclass)


def integrate_midpoint(x, y):
    """Calculate the integral using midpoint method.

    :param x: 1D array-like
    :param y: 1D array-like, with the same length as x.
    """
    _, res = cum_integrate_midpoint(x, y)
    return res[-1]


def cum_integrate_midpoint(x, y):
    """Calculate the cumulative integral using midpoint method.

    :param x: 1D array-like
    :param y: 1D array-like, with the same length as x."""
    x = np.array(x)
    y = np.array(y)
    dx = x[1:] - x[:-1]
    x_mid = 0.5 * (x[1:] + x[:-1])
    y_mid = 0.5 * (y[1:] + y[:-1])
    return x_mid, np.cumsum(dx * y_mid)



import appletree as apt

from appletree.component import ComponentSim


class NR(ComponentSim):
    norm_type = 'on_pdf'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.register(apt.plugins.common.FixedEnergySpectra)
        self.register(apt.plugins.common.PositionSpectra)
        self.register_all(apt.plugins.lyqy)
        self.register_all(apt.plugins.detector)
        self.register_all(apt.plugins.reconstruction)
        self.register_all(apt.plugins.efficiency)
